from __future__ import annotations

from typing import TYPE_CHECKING
from typing import Any

from conda_lock._vendor.cleo.application import Application
from conda_lock._vendor.cleo.commands.command import Command
from conda_lock._vendor.cleo.io.inputs.argument import Argument
from conda_lock._vendor.cleo.io.inputs.definition import Definition
from conda_lock._vendor.cleo.io.inputs.option import Option
from conda_lock._vendor.cleo.io.outputs.output import Type


if TYPE_CHECKING:
    from conda_lock._vendor.cleo.io.io import IO


class Descriptor:
    def describe(self, io: IO, obj: Any, **options: Any) -> None:
        self._io = io

        if isinstance(obj, Argument):
            self._describe_argument(obj, **options)
        elif isinstance(obj, Option):
            self._describe_option(obj, **options)
        elif isinstance(obj, Definition):
            self._describe_definition(obj, **options)
        elif isinstance(obj, Command):
            self._describe_command(obj, **options)
        elif isinstance(obj, Application):
            self._describe_application(obj, **options)

    def _write(self, content: str, decorated: bool = True) -> None:
        self._io.write(
            content, new_line=False, type=Type.NORMAL if decorated else Type.RAW
        )

    def _describe_argument(self, argument: Argument, **options: Any) -> None:
        raise NotImplementedError

    def _describe_option(self, option: Option, **options: Any) -> None:
        raise NotImplementedError

    def _describe_definition(self, definition: Definition, **options: Any) -> None:
        raise NotImplementedError

    def _describe_command(self, command: Command, **options: Any) -> None:
        raise NotImplementedError

    def _describe_application(self, application: Application, **options: Any) -> None:
        raise NotImplementedError



from typing import Optional, TypedDict


class FetchAction(TypedDict):
    """
    FETCH actions include all the entries from the corresponding package's
    repodata.json
    """

    channel: str
    constrains: Optional[list[str]]
    depends: Optional[list[str]]
    fn: str
    md5: str
    sha256: Optional[str]
    name: str
    subdir: str
    timestamp: int
    url: str
    version: str


class LinkAction(TypedDict):
    """
    LINK actions include only entries from conda-meta, notably missing
    dependency and constraint information
    """

    base_url: str
    channel: str
    dist_name: str
    name: str
    platform: str
    version: str


class InstallActions(TypedDict):
    LINK: list[LinkAction]
    FETCH: list[FetchAction]


class DryRunInstall(TypedDict):
    actions: InstallActions



"""Evaluate PEP 508 environment markers.

Environment markers are expressions such as `sys_platform == 'darwin'` that can
be attached to dependency specifications.

<https://www.python.org/dev/peps/pep-0508/#environment-markers>
"""

from typing import Union

from conda_lock.common import warn
from conda_lock.interfaces.vendored_poetry_markers import (
    AnyMarker,
    BaseMarker,
    EmptyMarker,
    MarkerUnion,
    MultiMarker,
    SingleMarker,
    parse_marker,
)
from conda_lock.pypi_solver import PlatformEnv


def get_names(marker: Union[BaseMarker, str]) -> set[str]:
    """Extract all environment marker names from a marker expression.

    >>> names = get_names(
    ...     "python_version < '3.9' and os_name == 'nt' or os_name == 'posix'"
    ... )
    >>> sorted(names)
    ['os_name', 'python_version']
    """
    if isinstance(marker, str):
        marker = parse_marker(marker)
    if isinstance(marker, SingleMarker):
        return {marker.name}
    if isinstance(marker, (MarkerUnion, MultiMarker)):
        return set.union(*[get_names(m) for m in marker.markers])
    if isinstance(marker, (AnyMarker, EmptyMarker)):
        return set()
    raise NotImplementedError(f"Unknown marker type: {marker!r}")


def evaluate_marker(marker: Union[BaseMarker, str, None], platform: str) -> bool:
    """Evaluate a marker expression for a given platform.

    This is intended to be used for parsing lock specifications, before the Python
    version is known, so markers like `python_version` are not supported.
    If the marker contains any unsupported names, a warning is issued, and the
    corresponding clause will evaluate to `True`.

    >>> evaluate_marker("sys_platform == 'darwin'", "osx-arm64")
    True
    >>> evaluate_marker("sys_platform == 'darwin'", "linux-64")
    False
    >>> evaluate_marker(None, "win-64")
    True

    # Unsupported names evaluate to True
    >>> evaluate_marker("python_version < '0' and implementation_name == 'q'", "win-64")
    True
    """
    if marker is None:
        return True
    if isinstance(marker, str):
        marker = parse_marker(marker)
    env = PlatformEnv(platform=platform)
    marker_env = env.get_marker_env()
    names = get_names(marker)
    supported_names = set(marker_env.keys())
    if not names <= supported_names:
        warn(
            f"Marker '{marker}' contains environment markers: "
            f"{names - supported_names}. Only {supported_names} are supported."
        )
    return marker.validate(marker_env)



"""Type definitions related to computing content hashes.

There is incidentally also a lot of virtual package stuff here.
"""

from typing import (
    Literal,
    Optional,
    Union,
)

from typing_extensions import NotRequired, TypeAlias, TypedDict


# Use TypeAlias to be descriptive about what kinds of strings are expected.
# This retains the flexibility of str, avoiding the need for awkward casting.
PlatformSubdirStr: TypeAlias = str
PackageNameStr: TypeAlias = str


class HashableVirtualPackage(TypedDict):
    """A dict that represents a fake package when computing the lockfile content hash"""

    name: PackageNameStr
    version: str
    build_string: str
    build_number: int
    build: str
    noarch: str
    depends: list[str]
    timestamp: int
    package_type: Optional[str]
    subdir: PlatformSubdirStr


class SerializedDependency(TypedDict):
    # _BaseDependency fields:
    name: str
    manager: Literal["conda", "pip"]
    category: str
    extras: list[str]
    markers: NotRequired[Optional[str]]
    # Note, markers was added in conda-lock v3

    # VersionedDependency fields:
    version: NotRequired[str]
    build: NotRequired[Optional[str]]
    conda_channel: NotRequired[Optional[str]]
    hash: NotRequired[Optional[str]]

    # URLDependency fields:
    url: NotRequired[str]
    hashes: NotRequired[list[str]]

    # VCSDependency fields:
    source: NotRequired[str]
    vcs: NotRequired[str]
    rev: NotRequired[Optional[str]]
    subdirectory: NotRequired[Optional[str]]

    # PathDependency fields:
    path: NotRequired[str]
    is_directory: NotRequired[bool]
    # # Also in VCSDependency:
    # subdirectory: NotRequired[Optional[str]]


class RepoMetadataInfo(TypedDict):
    subdir: PlatformSubdirStr


class SubdirMetadata(TypedDict):
    info: RepoMetadataInfo
    packages: dict[PackageNameStr, HashableVirtualPackage]


class EmptyDict(TypedDict):
    pass


HashableVirtualPackageRepresentation: "TypeAlias" = dict[
    PlatformSubdirStr, Union[SubdirMetadata, EmptyDict]
]


class SerializedLockspec(TypedDict):
    channels: list[str]
    specs: list[SerializedDependency]
    pip_repositories: NotRequired[list[str]]
    virtual_package_hash: NotRequired[HashableVirtualPackageRepresentation]



from collections import defaultdict
from typing import ClassVar, Optional

from conda_lock.lockfile.v1.models import (
    BaseLockedDependency,
    DependencySource,
    GitMeta,
    HashModel,
    InputMeta,
    LockKey,
    LockMeta,
    MetadataOption,
    TimeMeta,
)
from conda_lock.lockfile.v1.models import LockedDependency as LockedDependencyV1
from conda_lock.lockfile.v1.models import Lockfile as LockfileV1
from conda_lock.models import StrictModel


class LockedDependency(BaseLockedDependency):
    categories: set[str] = set()

    def to_v1(self) -> list[LockedDependencyV1]:
        """Convert a v2 dependency into a list of v1 dependencies.

        In case a v2 dependency might contain multiple categories, but a v1 dependency
        can only contain a single category, we represent multiple categories as a list
        of v1 dependencies that are identical except for the `category` field. The
        `category` field runs over all categories."""
        package_entries_per_category = [
            LockedDependencyV1(
                name=self.name,
                version=self.version,
                manager=self.manager,
                platform=self.platform,
                dependencies=self.dependencies,
                url=self.url,
                hash=self.hash,
                category=category,
                source=self.source,
                build=self.build,
                optional=category != "main",
            )
            for category in sorted(self.categories)
        ]
        return package_entries_per_category


class Lockfile(StrictModel):
    version: ClassVar[int] = 2

    package: list[LockedDependency]
    metadata: LockMeta

    def merge(self, other: "Optional[Lockfile]") -> "Lockfile":
        """
        merge self into other
        """
        if other is None:
            return self
        elif not isinstance(other, Lockfile):
            raise TypeError

        if self.metadata.channels != other.metadata.channels:
            raise ValueError(
                f"Cannot merge locked dependencies when the channels are not "
                f"consistent. {self.metadata.channels} != {other.metadata.channels}. "
                f"If the channels are indeed different, then you may need to delete "
                f"the existing lockfile and relock from scratch."
            )

        ours = {d.key(): d for d in self.package}
        theirs = {d.key(): d for d in other.package}

        # Pick ours preferentially
        package: list[LockedDependency] = []
        for key in sorted(set(ours.keys()).union(theirs.keys())):
            if key not in ours or key[-1] not in self.metadata.platforms:
                package.append(theirs[key])
            else:
                package.append(ours[key])

        # Resort the conda packages topologically
        final_package = self._toposort(package)
        return Lockfile(package=final_package, metadata=self.metadata | other.metadata)

    def toposort_inplace(self) -> None:
        self.package = self._toposort(self.package)

    def alphasort_inplace(self) -> None:
        # Sort the packages themselves by key (conda/pip, name, platform)
        self.package.sort(key=lambda d: d.key())
        for p in self.package:
            # Also ensure that the dependencies of each package are sorted
            # <https://github.com/conda/conda-lock/pull/654#issuecomment-2198453427>
            p.dependencies = {
                name: spec for name, spec in sorted(p.dependencies.items())
            }

    def filter_virtual_packages_inplace(self) -> None:
        self.package = [
            p
            for p in self.package
            if not (p.manager == "conda" and p.name.startswith("__"))
        ]

    @staticmethod
    def _toposort(package: list[LockedDependency]) -> list[LockedDependency]:
        platforms = {d.platform for d in package}

        # Resort the conda packages topologically
        final_package: list[LockedDependency] = []
        for platform in sorted(platforms):
            from conda_lock.interfaces.vendored_conda import toposort

            # Add the remaining non-conda packages in the order in which they appeared.
            # Order the pip packages topologically ordered (might be not 100% perfect if they depend on
            # other conda packages, but good enough
            for manager in ["conda", "pip"]:
                lookup = defaultdict(set)
                packages: dict[str, LockedDependency] = {}

                for d in package:
                    if d.platform != platform:
                        continue

                    if d.manager != manager:
                        continue

                    lookup[d.name] = set(d.dependencies)
                    packages[d.name] = d

                ordered = toposort(lookup)
                for package_name in ordered:
                    # since we could have a pure dep in here, that does not have a package
                    # eg a pip package that depends on a conda package (the conda package will not be in this list)
                    dep = packages.get(package_name)
                    if dep is None:
                        continue
                    if dep.manager != manager:
                        continue
                    final_package.append(dep)

        return final_package

    def to_v1(self) -> LockfileV1:
        # Each v2 package gives a list of v1 packages.
        # Flatten these into a single list of v1 packages.
        v1_packages = [
            package_entry_per_category
            for p in self.package
            for package_entry_per_category in p.to_v1()
        ]
        return LockfileV1(
            package=v1_packages,
            metadata=self.metadata,
        )


def _locked_dependency_v1_to_v2(
    package_entries_per_category: list[LockedDependencyV1],
) -> LockedDependency:
    """Convert a LockedDependency from v1 to v2.

    This is an inverse to `LockedDependency.to_v1()`.
    """
    # Dependencies are parsed from a v1 lockfile, so there will always be
    # at least one entry corresponding to what was parsed.
    assert len(package_entries_per_category) > 0
    # All the package entries should share the same key.
    assert all(
        d.key() == package_entries_per_category[0].key()
        for d in package_entries_per_category
    )

    categories = {d.category for d in package_entries_per_category}

    # Each entry should correspond to a distinct category
    assert len(categories) == len(package_entries_per_category)

    return LockedDependency(
        name=package_entries_per_category[0].name,
        version=package_entries_per_category[0].version,
        manager=package_entries_per_category[0].manager,
        platform=package_entries_per_category[0].platform,
        dependencies=package_entries_per_category[0].dependencies,
        url=package_entries_per_category[0].url,
        hash=package_entries_per_category[0].hash,
        categories=categories,
        source=package_entries_per_category[0].source,
        build=package_entries_per_category[0].build,
    )


def lockfile_v1_to_v2(lockfile_v1: LockfileV1) -> Lockfile:
    """Convert a Lockfile from v1 to v2.

    Entries may share the same key if they represent a dependency
    belonging to multiple categories. They must be collected here.
    """
    dependencies_for_key: dict[LockKey, list[LockedDependencyV1]] = defaultdict(list)
    for dep in lockfile_v1.package:
        dependencies_for_key[dep.key()].append(dep)

    v2_packages = [
        _locked_dependency_v1_to_v2(package_entries_per_category)
        for package_entries_per_category in dependencies_for_key.values()
    ]

    return Lockfile(
        package=v2_packages,
        metadata=lockfile_v1.metadata,
    )


class UpdateSpecification:
    def __init__(
        self,
        locked: Optional[list[LockedDependency]] = None,
        update: Optional[list[str]] = None,
    ):
        self.locked = locked or []
        self.update = update or []


__all__ = [
    "DependencySource",
    "GitMeta",
    "HashModel",
    "InputMeta",
    "LockMeta",
    "LockedDependency",
    "Lockfile",
    "MetadataOption",
    "TimeMeta",
    "UpdateSpecification",
    "lockfile_v1_to_v2",
]



from __future__ import annotations

from conda_lock._vendor.poetry.installation.operations.install import Install
from conda_lock._vendor.poetry.installation.operations.uninstall import Uninstall
from conda_lock._vendor.poetry.installation.operations.update import Update


__all__ = ["Install", "Uninstall", "Update"]



from __future__ import annotations

import tempfile

from pathlib import Path
from typing import TYPE_CHECKING

from conda_lock._vendor.poetry.core.utils.helpers import temporary_directory

from conda_lock._vendor.poetry.utils.helpers import extractall
from conda_lock._vendor.poetry.utils.isolated_build import isolated_builder


if TYPE_CHECKING:
    from build import DistributionType

    from conda_lock._vendor.poetry.repositories import RepositoryPool
    from conda_lock._vendor.poetry.utils.cache import ArtifactCache
    from conda_lock._vendor.poetry.utils.env import Env


class ChefError(Exception): ...


class Chef:
    def __init__(
        self, artifact_cache: ArtifactCache, env: Env, pool: RepositoryPool
    ) -> None:
        self._env = env
        self._pool = pool
        self._artifact_cache = artifact_cache

    def prepare(
        self, archive: Path, output_dir: Path | None = None, *, editable: bool = False
    ) -> Path:
        if not self._should_prepare(archive):
            return archive

        if archive.is_dir():
            destination = output_dir or Path(tempfile.mkdtemp(prefix="poetry-chef-"))
            return self._prepare(archive, destination=destination, editable=editable)

        return self._prepare_sdist(archive, destination=output_dir)

    def _prepare(
        self, directory: Path, destination: Path, *, editable: bool = False
    ) -> Path:
        distribution: DistributionType = "editable" if editable else "wheel"
        with isolated_builder(
            source=directory,
            distribution=distribution,
            python_executable=self._env.python,
            pool=self._pool,
        ) as builder:
            return Path(
                builder.build(
                    distribution,
                    destination.as_posix(),
                )
            )

    def _prepare_sdist(self, archive: Path, destination: Path | None = None) -> Path:
        from conda_lock._vendor.poetry.core.packages.utils.link import Link

        suffix = archive.suffix
        zip = suffix == ".zip"

        with temporary_directory() as tmp_dir:
            archive_dir = Path(tmp_dir)
            extractall(source=archive, dest=archive_dir, zip=zip)

            elements = list(archive_dir.glob("*"))

            if len(elements) == 1 and elements[0].is_dir():
                sdist_dir = elements[0]
            else:
                sdist_dir = archive_dir / archive.name.rstrip(suffix)
                if not sdist_dir.is_dir():
                    sdist_dir = archive_dir

            if destination is None:
                destination = self._artifact_cache.get_cache_directory_for_link(
                    Link(archive.as_uri())
                )

            destination.mkdir(parents=True, exist_ok=True)

            return self._prepare(
                sdist_dir,
                destination,
            )

    def _should_prepare(self, archive: Path) -> bool:
        return archive.is_dir() or not self._is_wheel(archive)

    @classmethod
    def _is_wheel(cls, archive: Path) -> bool:
        return archive.suffix == ".whl"



# Copyright (C) 2012 Anaconda, Inc
# SPDX-License-Identifier: BSD-3-Clause
"""CLI implementation for `conda list`.

Lists all packages installed into an environment.
"""

import logging
import re
from argparse import ArgumentParser, Namespace, _SubParsersAction
from os.path import isdir, isfile

log = logging.getLogger(__name__)


def configure_parser(sub_parsers: _SubParsersAction, **kwargs) -> ArgumentParser:
    from ..auxlib.ish import dals
    from .helpers import (
        add_parser_json,
        add_parser_prefix,
        add_parser_show_channel_urls,
    )

    summary = "List installed packages in a conda environment."
    description = summary
    epilog = dals(
        """
        Examples:

        List all packages in the current environment::

            conda list

        List all packages in reverse order::

            conda list --reverse

        List all packages installed into the environment 'myenv'::

            conda list -n myenv

        List all packages that begin with the letters "py", using regex::

            conda list ^py

        Save packages for future use::

            conda list --export > package-list.txt

        Reinstall packages from an export file::

            conda create -n myenv --file package-list.txt

        """
    )

    p = sub_parsers.add_parser(
        "list",
        help=summary,
        description=description,
        epilog=epilog,
        **kwargs,
    )
    add_parser_prefix(p)
    add_parser_json(p)
    add_parser_show_channel_urls(p)
    p.add_argument(
        "--reverse",
        action="store_true",
        default=False,
        help="List installed packages in reverse order.",
    )
    p.add_argument(
        "-c",
        "--canonical",
        action="store_true",
        help="Output canonical names of packages only.",
    )
    p.add_argument(
        "-f",
        "--full-name",
        action="store_true",
        help="Only search for full names, i.e., ^<regex>$. "
        "--full-name NAME is identical to regex '^NAME$'.",
    )
    p.add_argument(
        "--explicit",
        action="store_true",
        help="List explicitly all installed conda packages with URL "
        "(output may be used by conda create --file).",
    )
    p.add_argument(
        "--md5",
        action="store_true",
        help="Add MD5 hashsum when using --explicit.",
    )
    p.add_argument(
        "-e",
        "--export",
        action="store_true",
        help="Output explicit, machine-readable requirement strings instead of "
        "human-readable lists of packages. This output may be used by "
        "conda create --file.",
    )
    p.add_argument(
        "-r",
        "--revisions",
        action="store_true",
        help="List the revision history.",
    )
    p.add_argument(
        "--no-pip",
        action="store_false",
        default=True,
        dest="pip",
        help="Do not include pip-only installed packages.",
    )
    p.add_argument(
        "--auth",
        action="store_false",
        default=True,
        dest="remove_auth",
        help="In explicit mode, leave authentication details in package URLs. "
        "They are removed by default otherwise.",
    )
    p.add_argument(
        "regex",
        action="store",
        nargs="?",
        help="List only packages matching this regular expression.",
    )
    p.set_defaults(func="conda.cli.main_list.execute")

    return p


def print_export_header(subdir):
    print("# This file may be used to create an environment using:")
    print("# $ conda create --name <env> --file <this file>")
    print(f"# platform: {subdir}")


def get_packages(installed, regex):
    pat = re.compile(regex, re.I) if regex else None
    for prefix_rec in sorted(installed, key=lambda x: x.name.lower()):
        if pat and pat.search(prefix_rec.name) is None:
            continue
        yield prefix_rec


def list_packages(
    prefix,
    regex=None,
    format="human",
    reverse=False,
    show_channel_urls=None,
):
    from ..base.constants import DEFAULTS_CHANNEL_NAME
    from ..base.context import context
    from ..core.prefix_data import PrefixData
    from .common import disp_features

    res = 0

    installed = sorted(
        PrefixData(prefix, pip_interop_enabled=True).iter_records(),
        key=lambda x: x.name,
    )

    packages = []
    for prec in get_packages(installed, regex) if regex else installed:
        if format == "canonical":
            packages.append(
                prec.dist_fields_dump() if context.json else prec.dist_str()
            )
            continue
        if format == "export":
            packages.append("=".join((prec.name, prec.version, prec.build)))
            continue

        features = set(prec.get("features") or ())
        disp = "%(name)-25s %(version)-15s %(build)15s" % prec
        disp += f"  {disp_features(features)}"
        schannel = prec.get("schannel")
        show_channel_urls = show_channel_urls or context.show_channel_urls
        if (
            show_channel_urls
            or show_channel_urls is None
            and schannel != DEFAULTS_CHANNEL_NAME
        ):
            disp += f"  {schannel}"

        packages.append(disp)

    if reverse:
        packages = reversed(packages)

    result = []
    if format == "human":
        result = [
            f"# packages in environment at {prefix}:",
            "#",
            "# %-23s %-15s %15s  Channel" % ("Name", "Version", "Build"),
        ]
    result.extend(packages)

    return res, result


def print_packages(
    prefix,
    regex=None,
    format="human",
    reverse=False,
    piplist=False,
    json=False,
    show_channel_urls=None,
):
    from ..base.context import context
    from .common import stdout_json

    if not isdir(prefix):
        from ..exceptions import EnvironmentLocationNotFound

        raise EnvironmentLocationNotFound(prefix)

    if not json:
        if format == "export":
            print_export_header(context.subdir)

    exitcode, output = list_packages(
        prefix,
        regex,
        format=format,
        reverse=reverse,
        show_channel_urls=show_channel_urls,
    )
    if context.json:
        stdout_json(output)

    else:
        print("\n".join(map(str, output)))

    return exitcode


def print_explicit(prefix, add_md5=False, remove_auth=True):
    from ..base.constants import UNKNOWN_CHANNEL
    from ..base.context import context
    from ..common import url as common_url
    from ..core.prefix_data import PrefixData

    if not isdir(prefix):
        from ..exceptions import EnvironmentLocationNotFound

        raise EnvironmentLocationNotFound(prefix)
    print_export_header(context.subdir)
    print("@EXPLICIT")
    for prefix_record in PrefixData(prefix).iter_records_sorted():
        url = prefix_record.get("url")
        if not url or url.startswith(UNKNOWN_CHANNEL):
            print("# no URL for: {}".format(prefix_record["fn"]))
            continue
        if remove_auth:
            url = common_url.remove_auth(common_url.split_anaconda_token(url)[0])
        md5 = prefix_record.get("md5")
        print(url + (f"#{md5}" if add_md5 and md5 else ""))


def execute(args: Namespace, parser: ArgumentParser) -> int:
    from ..base.context import context
    from ..gateways.disk.test import is_conda_environment
    from ..history import History
    from .common import stdout_json

    prefix = context.target_prefix
    if not is_conda_environment(prefix):
        from ..exceptions import EnvironmentLocationNotFound

        raise EnvironmentLocationNotFound(prefix)

    regex = args.regex
    if args.full_name:
        regex = rf"^{regex}$"

    if args.revisions:
        h = History(prefix)
        if isfile(h.path):
            if not context.json:
                h.print_log()
            else:
                stdout_json(h.object_log())
        else:
            from ..exceptions import PathNotFoundError

            raise PathNotFoundError(h.path)
        return 0

    if args.explicit:
        print_explicit(prefix, args.md5, args.remove_auth)
        return 0

    if args.canonical:
        format = "canonical"
    elif args.export:
        format = "export"
    else:
        format = "human"

    if context.json:
        format = "canonical"

    return print_packages(
        prefix,
        regex,
        format,
        reverse=args.reverse,
        piplist=args.pip,
        json=context.json,
        show_channel_urls=context.show_channel_urls,
    )



from __future__ import annotations

from conda_lock._vendor.poetry.repositories.repository import Repository
from conda_lock._vendor.poetry.repositories.repository_pool import RepositoryPool


__all__ = ["Repository", "RepositoryPool"]



from conda_lock._vendor.poetry.config.config import Config
from conda_lock._vendor.poetry.core.constraints.version.version_constraint import (
    VersionConstraint,
)
from conda_lock._vendor.poetry.core.packages.dependency import (
    Dependency as PoetryDependency,
)
from conda_lock._vendor.poetry.core.packages.directory_dependency import (
    DirectoryDependency as PoetryDirectoryDependency,
)
from conda_lock._vendor.poetry.core.packages.file_dependency import (
    FileDependency as PoetryFileDependency,
)
from conda_lock._vendor.poetry.core.packages.package import Package as PoetryPackage
from conda_lock._vendor.poetry.core.packages.project_package import (
    ProjectPackage as PoetryProjectPackage,
)
from conda_lock._vendor.poetry.core.packages.url_dependency import (
    URLDependency as PoetryURLDependency,
)
from conda_lock._vendor.poetry.core.packages.utils.link import Link
from conda_lock._vendor.poetry.core.packages.vcs_dependency import (
    VCSDependency as PoetryVCSDependency,
)
from conda_lock._vendor.poetry.factory import Factory
from conda_lock._vendor.poetry.installation.chooser import Chooser
from conda_lock._vendor.poetry.installation.operations.operation import Operation
from conda_lock._vendor.poetry.puzzle import Solver as PoetrySolver
from conda_lock._vendor.poetry.repositories.pypi_repository import PyPiRepository
from conda_lock._vendor.poetry.repositories.repository import Repository
from conda_lock._vendor.poetry.repositories.repository_pool import (
    RepositoryPool as Pool,
)
from conda_lock._vendor.poetry.utils.env import Env, VirtualEnv


__all__ = [
    "Chooser",
    "Config",
    "Env",
    "Factory",
    "Link",
    "Operation",
    "PoetryDependency",
    "PoetryDirectoryDependency",
    "PoetryFileDependency",
    "PoetryPackage",
    "PoetryProjectPackage",
    "PoetrySolver",
    "PoetryURLDependency",
    "PoetryVCSDependency",
    "Pool",
    "PyPiRepository",
    "Repository",
    "VersionConstraint",
    "VirtualEnv",
]



from __future__ import annotations

from conda_lock._vendor.poetry.packages.dependency_package import DependencyPackage
from conda_lock._vendor.poetry.packages.locker import Locker
from conda_lock._vendor.poetry.packages.package_collection import PackageCollection


__all__ = ["DependencyPackage", "Locker", "PackageCollection"]



from __future__ import annotations

from pathlib import Path
from typing import TYPE_CHECKING

from conda_lock._vendor.poetry.core.packages.dependency import Dependency
from conda_lock._vendor.poetry.core.packages.project_package import ProjectPackage

from conda_lock._vendor.poetry.__version__ import __version__
from conda_lock._vendor.poetry.console.commands.installer_command import InstallerCommand
from conda_lock._vendor.poetry.factory import Factory
from conda_lock._vendor.poetry.pyproject.toml import PyProjectTOML
from conda_lock._vendor.poetry.utils.env import EnvManager
from conda_lock._vendor.poetry.utils.env import SystemEnv
from conda_lock._vendor.poetry.utils.helpers import directory


if TYPE_CHECKING:
    from conda_lock._vendor.poetry.poetry import Poetry
    from conda_lock._vendor.poetry.utils.env import Env


class SelfCommand(InstallerCommand):
    ADDITIONAL_PACKAGE_GROUP = "additional"

    @staticmethod
    def get_default_system_pyproject_file() -> Path:
        # We separate this out to avoid unwanted side effect during testing while
        # maintaining dynamic use in help text.
        #
        # This is not ideal, but is the simplest solution for now.
        from conda_lock._vendor.poetry.locations import CONFIG_DIR

        return Path(CONFIG_DIR).joinpath("pyproject.toml")

    @property
    def system_pyproject(self) -> Path:
        file = self.get_default_system_pyproject_file()
        file.parent.mkdir(parents=True, exist_ok=True)
        return file

    def reset_env(self) -> None:
        self._env = EnvManager.get_system_env(naive=True)

    @property
    def env(self) -> Env:
        if not isinstance(self._env, SystemEnv):
            self.reset_env()
        assert self._env is not None
        return self._env

    @property
    def default_group(self) -> str:
        return self.ADDITIONAL_PACKAGE_GROUP

    @property
    def activated_groups(self) -> set[str]:
        return {self.default_group}

    def generate_system_pyproject(self) -> None:
        preserved = {}

        if self.system_pyproject.exists():
            content = PyProjectTOML(self.system_pyproject).poetry_config

            for key in {"group", "source"}:
                if key in content:
                    preserved[key] = content[key]

        package = ProjectPackage(name="poetry-instance", version=__version__)
        package.add_dependency(Dependency(name="poetry", constraint=f"{__version__}"))

        package.python_versions = ".".join(str(v) for v in self.env.version_info[:3])

        content = Factory.create_pyproject_from_package(package=package)
        content["tool"]["poetry"]["package-mode"] = False  # type: ignore[index]

        for key in preserved:
            content["tool"]["poetry"][key] = preserved[key]  # type: ignore[index]

        pyproject = PyProjectTOML(self.system_pyproject)
        pyproject.file.write(content)

    def reset_poetry(self) -> None:
        with directory(self.system_pyproject.parent):
            self.generate_system_pyproject()
            self._poetry = Factory().create_poetry(
                self.system_pyproject.parent, io=self.io, disable_plugins=True
            )

    @property
    def poetry(self) -> Poetry:
        if self._poetry is None:
            self.reset_poetry()

        assert self._poetry is not None
        return self._poetry

    def _system_project_handle(self) -> int:
        """
        This is a helper method that by default calls the handle method implemented in
        the child class's next MRO sibling. Override this if you want special handling
        either before calling the handle() from the super class or have custom logic
        to handle the command.

        The default implementations handles cases where a `self` command delegates
        handling to an existing command. Eg: `SelfAddCommand(SelfCommand, AddCommand)`.
        """
        return_code: int = super().handle()
        return return_code

    def reset(self) -> None:
        """
        Reset current command instance's environment and poetry instances to ensure
        use of the system specific ones.
        """
        self.reset_env()
        self.reset_poetry()

    def handle(self) -> int:
        # We override the base class's handle() method to ensure that poetry and env
        # are reset to work within the system project instead of current context.
        # Further, during execution, the working directory is temporarily changed
        # to parent directory of Poetry system pyproject.toml file.
        #
        # This method **should not** be overridden in child classes as it may have
        # unexpected consequences.

        self.reset()

        with directory(self.system_pyproject.parent):
            return self._system_project_handle()



from __future__ import annotations

import functools
import time

from collections import defaultdict
from contextlib import contextmanager
from typing import TYPE_CHECKING

from conda_lock._vendor.poetry.core.version.markers import AnyMarker
from conda_lock._vendor.poetry.core.version.markers import EmptyMarker
from conda_lock._vendor.poetry.core.version.markers import parse_marker

from conda_lock._vendor.poetry.mixology import resolve_version
from conda_lock._vendor.poetry.mixology.failure import SolveFailureError
from conda_lock._vendor.poetry.packages.transitive_package_info import TransitivePackageInfo
from conda_lock._vendor.poetry.puzzle.exceptions import OverrideNeededError
from conda_lock._vendor.poetry.puzzle.exceptions import SolverProblemError
from conda_lock._vendor.poetry.puzzle.provider import Indicator
from conda_lock._vendor.poetry.puzzle.provider import Provider


if TYPE_CHECKING:
    from collections.abc import Collection
    from collections.abc import Iterator
    from collections.abc import Sequence

    from conda_lock._vendor.cleo.io.io import IO
    from packaging.utils import NormalizedName
    from conda_lock._vendor.poetry.core.constraints.version import VersionConstraint
    from conda_lock._vendor.poetry.core.packages.dependency import Dependency
    from conda_lock._vendor.poetry.core.packages.package import Package
    from conda_lock._vendor.poetry.core.packages.project_package import ProjectPackage
    from conda_lock._vendor.poetry.core.version.markers import BaseMarker
    from typing_extensions import Self

    from conda_lock._vendor.poetry.puzzle.transaction import Transaction
    from conda_lock._vendor.poetry.repositories import RepositoryPool
    from conda_lock._vendor.poetry.utils.env import Env

    MarkerOriginDict = defaultdict[Package, defaultdict[Package, BaseMarker]]


class Solver:
    def __init__(
        self,
        package: ProjectPackage,
        pool: RepositoryPool,
        installed: list[Package],
        locked: list[Package],
        io: IO,
        active_root_extras: Collection[NormalizedName] | None = None,
    ) -> None:
        self._package = package
        self._pool = pool
        self._installed_packages = installed
        self._locked_packages = locked
        self._io = io

        self._provider = Provider(
            self._package,
            self._pool,
            self._io,
            locked=locked,
            active_root_extras=active_root_extras,
        )
        self._overrides: list[dict[Package, dict[str, Dependency]]] = []

    @property
    def provider(self) -> Provider:
        return self._provider

    @contextmanager
    def use_environment(self, env: Env) -> Iterator[None]:
        with self.provider.use_environment(env):
            yield

    def solve(
        self, use_latest: Collection[NormalizedName] | None = None
    ) -> Transaction:
        from conda_lock._vendor.poetry.puzzle.transaction import Transaction

        with self._progress(), self._provider.use_latest_for(use_latest or []):
            start = time.time()
            packages = self._solve()
            # simplify markers by removing redundant information
            for transitive_info in packages.values():
                for group, marker in transitive_info.markers.items():
                    transitive_info.markers[group] = simplify_marker(
                        marker, self._package.python_constraint
                    )
            end = time.time()

            if len(self._overrides) > 1:
                self._provider.debug(
                    # ignore the warning as provider does not do interpolation
                    f"Complete version solving took {end - start:.3f}"
                    f" seconds with {len(self._overrides)} overrides"
                )
                self._provider.debug(
                    # ignore the warning as provider does not do interpolation
                    "Resolved with overrides:"
                    f" {', '.join(f'({b})' for b in self._overrides)}"
                )

        for p in packages:
            if p.yanked:
                message = (
                    f"The locked version {p.pretty_version} for {p.pretty_name} is a"
                    " yanked version."
                )
                if p.yanked_reason:
                    message += f" Reason for being yanked: {p.yanked_reason}"
                self._io.write_error_line(f"<warning>Warning: {message}</warning>")

        return Transaction(
            self._locked_packages,
            packages,
            installed_packages=self._installed_packages,
            root_package=self._package,
        )

    @contextmanager
    def _progress(self) -> Iterator[None]:
        if not self._io.output.is_decorated() or self._provider.is_debugging():
            self._io.write_line("Resolving dependencies...")
            yield
        else:
            indicator = Indicator(
                self._io, "{message}{context}<debug>({elapsed:2s})</debug>"
            )

            with indicator.auto(
                "<info>Resolving dependencies...</info>",
                "<info>Resolving dependencies...</info>",
            ):
                yield

    def _solve_in_compatibility_mode(
        self,
        overrides: tuple[dict[Package, dict[str, Dependency]], ...],
    ) -> dict[Package, TransitivePackageInfo]:
        override_packages: list[
            tuple[
                dict[Package, dict[str, Dependency]],
                dict[Package, TransitivePackageInfo],
            ]
        ] = []
        for override in overrides:
            self._provider.debug(
                # ignore the warning as provider does not do interpolation
                "<comment>Retrying dependency resolution "
                f"with the following overrides ({override}).</comment>"
            )
            self._provider.set_overrides(override)
            new_packages = self._solve()
            override_packages.append((override, new_packages))

        return merge_override_packages(override_packages)

    def _solve(self) -> dict[Package, TransitivePackageInfo]:
        if self._provider._overrides:
            self._overrides.append(self._provider._overrides)

        try:
            result = resolve_version(self._package, self._provider)

            packages = result.packages
        except OverrideNeededError as e:
            return self._solve_in_compatibility_mode(e.overrides)
        except SolveFailureError as e:
            raise SolverProblemError(e)

        return self._aggregate_solved_packages(packages)

    def _aggregate_solved_packages(
        self, packages: list[Package]
    ) -> dict[Package, TransitivePackageInfo]:
        combined_nodes, markers = depth_first_search(
            PackageNode(self._package, packages)
        )
        results = dict(aggregate_package_nodes(nodes) for nodes in combined_nodes)
        calculate_markers(results, markers)

        # Merging feature packages with base packages
        solved_packages = {}
        for package in packages:
            if package.features:
                for _package in packages:
                    if (
                        not _package.features
                        and _package.name == package.name
                        and _package.version == package.version
                    ):
                        for dep in package.requires:
                            # Prevent adding base package as a dependency to itself
                            if _package.name == dep.name:
                                continue

                            # Avoid duplication.
                            if any(
                                _dep == dep and _dep.marker == dep.marker
                                for _dep in _package.requires
                            ):
                                continue

                            _package.add_dependency(dep)
            else:
                solved_packages[package] = results[package]

        return solved_packages


DFSNodeID = tuple[str, frozenset[str], bool]


class DFSNode:
    def __init__(self, id: DFSNodeID, name: str, base_name: str) -> None:
        self.id = id
        self.name = name
        self.base_name = base_name

    def reachable(self) -> Sequence[Self]:
        return []

    def visit(self, parents: list[PackageNode]) -> None:
        pass

    def __str__(self) -> str:
        return str(self.id)


def depth_first_search(
    source: PackageNode,
) -> tuple[list[list[PackageNode]], MarkerOriginDict]:
    back_edges: dict[DFSNodeID, list[PackageNode]] = defaultdict(list)
    markers: MarkerOriginDict = defaultdict(lambda: defaultdict(EmptyMarker))
    visited: set[DFSNodeID] = set()
    topo_sorted_nodes: list[PackageNode] = []

    dfs_visit(source, back_edges, visited, topo_sorted_nodes, markers)

    # Combine the nodes by name
    combined_nodes: dict[str, list[PackageNode]] = defaultdict(list)
    for node in topo_sorted_nodes:
        node.visit(back_edges[node.id])
        combined_nodes[node.name].append(node)

    combined_topo_sorted_nodes: list[list[PackageNode]] = [
        combined_nodes.pop(node.name)
        for node in topo_sorted_nodes
        if node.name in combined_nodes
    ]

    return combined_topo_sorted_nodes, markers


def dfs_visit(
    node: PackageNode,
    back_edges: dict[DFSNodeID, list[PackageNode]],
    visited: set[DFSNodeID],
    sorted_nodes: list[PackageNode],
    markers: MarkerOriginDict,
) -> None:
    if node.id in visited:
        return
    visited.add(node.id)

    for out_neighbor in node.reachable():
        back_edges[out_neighbor.id].append(node)
        marker = markers[out_neighbor.package][node.package]
        markers[out_neighbor.package][node.package] = marker.union(
            out_neighbor.marker
            if node.package.is_root()
            else out_neighbor.marker.without_extras()
        )
        dfs_visit(out_neighbor, back_edges, visited, sorted_nodes, markers)
    sorted_nodes.insert(0, node)


class PackageNode(DFSNode):
    def __init__(
        self,
        package: Package,
        packages: list[Package],
        previous: PackageNode | None = None,
        dep: Dependency | None = None,
        marker: BaseMarker | None = None,
    ) -> None:
        self.package = package
        self.packages = packages

        self.dep = dep
        self.marker = marker or AnyMarker()
        self.depth = -1

        if not previous:
            self.groups: frozenset[str] = frozenset()
            self.optional = True
        elif dep:
            self.groups = dep.groups
            self.optional = dep.is_optional()
        else:
            raise ValueError("Both previous and dep must be passed")

        package_repr = repr(package)
        super().__init__(
            (package_repr, self.groups, self.optional),
            package_repr,
            package.name,
        )

    def reachable(self) -> Sequence[PackageNode]:
        children: list[PackageNode] = []

        for dependency in self.package.all_requires:
            for pkg in self.packages:
                if pkg.complete_name == dependency.complete_name and pkg.satisfies(
                    dependency
                ):
                    marker = dependency.marker
                    if self.package.is_root() and dependency.in_extras:
                        marker = marker.intersect(
                            parse_marker(
                                " or ".join(
                                    f'extra == "{extra}"'
                                    for extra in dependency.in_extras
                                )
                            )
                        )
                    children.append(
                        PackageNode(
                            pkg,
                            self.packages,
                            self,
                            self.dep or dependency,
                            marker,
                        )
                    )

        return children

    def visit(self, parents: list[PackageNode]) -> None:
        # The root package, which has no parents, is defined as having depth -1
        # So that the root package's top-level dependencies have depth 0.
        self.depth = 1 + max(
            [
                parent.depth if parent.base_name != self.base_name else parent.depth - 1
                for parent in parents
            ]
            + [-2]
        )


def aggregate_package_nodes(
    nodes: list[PackageNode],
) -> tuple[Package, TransitivePackageInfo]:
    package = nodes[0].package
    depth = max(node.depth for node in nodes)
    groups: set[str] = set()
    for node in nodes:
        groups.update(node.groups)

    optional = all(node.optional for node in nodes)
    for node in nodes:
        node.depth = depth
        node.optional = optional

    package.optional = optional

    # TransitivePackageInfo.markers is updated later,
    # because the nodes of all packages have to be aggregated first.
    return package, TransitivePackageInfo(depth, groups, {})


def calculate_markers(
    packages: dict[Package, TransitivePackageInfo], markers: MarkerOriginDict
) -> None:
    # group packages by depth
    packages_by_depth: dict[int, list[Package]] = defaultdict(list)
    max_depth = -1
    for package, info in packages.items():
        max_depth = max(max_depth, info.depth)
        packages_by_depth[info.depth].append(package)

    # calculate markers from lowest to highest depth
    # (start with depth 0 because the root package has depth -1)
    has_incomplete_markers = True
    while has_incomplete_markers:
        has_incomplete_markers = False
        for depth in range(max_depth + 1):
            for package in packages_by_depth[depth]:
                transitive_info = packages[package]
                transitive_marker: dict[str, BaseMarker] = {
                    group: EmptyMarker() for group in transitive_info.groups
                }
                for parent, m in markers[package].items():
                    parent_info = packages[parent]
                    if parent_info.groups:
                        if parent_info.groups != set(parent_info.markers):
                            # there is a cycle -> we need one more iteration
                            has_incomplete_markers = True
                            continue
                        for group in parent_info.groups:
                            transitive_marker[group] = transitive_marker[group].union(
                                parent_info.markers[group].intersect(m)
                            )
                    else:
                        for group in transitive_info.groups:
                            transitive_marker[group] = transitive_marker[group].union(m)
                transitive_info.markers = transitive_marker


def merge_override_packages(
    override_packages: list[
        tuple[
            dict[Package, dict[str, Dependency]], dict[Package, TransitivePackageInfo]
        ]
    ],
) -> dict[Package, TransitivePackageInfo]:
    result: dict[Package, TransitivePackageInfo] = {}
    all_packages: dict[
        Package, list[tuple[Package, TransitivePackageInfo, BaseMarker]]
    ] = {}
    for override, o_packages in override_packages:
        override_marker: BaseMarker = AnyMarker()
        for deps in override.values():
            for dep in deps.values():
                override_marker = override_marker.intersect(dep.marker.without_extras())
        for package, info in o_packages.items():
            all_packages.setdefault(package, []).append(
                (package, info, override_marker)
            )
    for package_duplicates in all_packages.values():
        base = package_duplicates[0]
        package = base[0]
        package_info = base[1]
        first_override_marker = base[2]
        result[package] = package_info
        package_info.depth = max(info.depth for _, info, _ in package_duplicates)
        package_info.groups = {
            g for _, info, _ in package_duplicates for g in info.groups
        }
        if all(
            info.markers == package_info.markers for _, info, _ in package_duplicates
        ):
            # performance shortcut:
            # if markers are the same for all overrides,
            # we can use less expensive marker operations
            override_marker = EmptyMarker()
            for _, _, marker in package_duplicates:
                override_marker = override_marker.union(marker)
            package_info.markers = {
                group: override_marker.intersect(marker)
                for group, marker in package_info.markers.items()
            }
        else:
            # fallback / general algorithm with performance issues
            for group, marker in package_info.markers.items():
                package_info.markers[group] = first_override_marker.intersect(marker)
            for _, info, override_marker in package_duplicates[1:]:
                for group, marker in info.markers.items():
                    package_info.markers[group] = package_info.markers.get(
                        group, EmptyMarker()
                    ).union(override_marker.intersect(marker))
        for duplicate_package, _, _ in package_duplicates[1:]:
            for dep in duplicate_package.requires:
                if dep not in package.requires:
                    package.add_dependency(dep)
    return result


@functools.cache
def simplify_marker(
    marker: BaseMarker, python_constraint: VersionConstraint
) -> BaseMarker:
    """
    Remove constraints from markers that are covered by the projects Python constraint.

    Use cache because we call this function often for the same markers.
    """
    return marker.reduce_by_python_constraint(python_constraint)



"""
Conda lock supports two kinds of credentials used for channels:

## Token based

These are used by anaconda.org, Anaconda Enterprise, and Quetz.
To pass one of these channels, specify them in your source with an environment variable.
Make sure this environment variable is not expanded.

Example:
--channel 'http://host.com/t/$MY_REPO_TOKEN/channel'
# TODO: Detect environment variables that match a channel specified incorrectly.

## Simple Auth

For other channels (such as those self-managed), you may be using standard
username/password auth:

Example:
--channel 'http://$USER:$PASSWORD@host.com/channel'

## What gets stored

Since credential parts are both volatile and secret, conda-lock will not store
the raw version of a URL. If it encounters a channel URL that contains credentials,
it will search the available environment variables for a match. When found, that portion
of the URL will be replaced with an environment variable.

Since conda also performs env var substitution, the rendered output can contain env vars
which will be handled correctly by conda/mamba.
"""

import copy
import logging
import os
import re

from posixpath import expandvars
from typing import Any, Optional
from urllib.parse import unquote, urlparse, urlunparse

from pydantic import BaseModel, ConfigDict

from conda_lock._vendor.conda.common.url import (
    mask_anaconda_token,
    split_anaconda_token,
)


logger = logging.getLogger(__name__)
token_pattern = re.compile(r"(.*)(/t/\$?\{?[a-zA-Z0-9-_]*\}?)(/.*)")


class _CondaUrl(BaseModel):
    """A high-level representation of a URL that may contain credentials.

    This is an intermediate representation that is used after parsing but before
    the URL is used to create a Channel object.
    """

    raw_url: str
    env_var_url: str
    token: Optional[str] = None
    token_env_var: Optional[str] = None
    user: Optional[str] = None
    user_env_var: Optional[str] = None
    password: Optional[str] = None
    password_env_var: Optional[str] = None


class Channel(BaseModel):
    model_config = ConfigDict(frozen=True)
    url: str
    used_env_vars: tuple[str, ...]

    @classmethod
    def from_string(cls, value: str) -> "Channel":
        """The primary constructor for Channel.

        >>> Channel.from_string("conda-forge")
        Channel(url='conda-forge')

        Channels can be specified with a label
        >>> Channel.from_string("conda-forge/label/micromamba_prerelease")
        Channel(url='conda-forge/label/micromamba_prerelease')

        Channels can contain tokens, and these will be replaced with env vars
        >>> os.environ["MY_A_REPO_TOKEN"] = "tk-123-456"
        >>> Channel.from_string(
        ...     "https://host.com/t/tk-123-456/channel"
        ... )  # doctest: +NORMALIZE_WHITESPACE
        Channel(url='https://host.com/t/${MY_A_REPO_TOKEN}/channel',
            used_env_vars=('MY_A_REPO_TOKEN',))

        Channels can contain username/password credentials
        >>> os.environ["MY_A_USERNAME"] = "user"
        >>> os.environ["MY_A_PASSWORD"] = "pass"
        >>> Channel.from_string(
        ...     "https://user:pass@host.com/channel"
        ... )  # doctest: +NORMALIZE_WHITESPACE
        Channel(url='https://${MY_A_USERNAME}:${MY_A_PASSWORD}@host.com/channel',
            used_env_vars=('MY_A_PASSWORD', 'MY_A_USERNAME'))

        >>> del os.environ["MY_A_USERNAME"]
        >>> del os.environ["MY_A_PASSWORD"]
        >>> del os.environ["MY_A_REPO_TOKEN"]
        """
        if "://" in value:
            conda_url = _conda_url_from_string(value)
            channel = _channel_from_conda_url(conda_url)
            return channel
        return cls(url=value, used_env_vars=())

    def env_replaced_url(self) -> str:
        """Expand environment variables in the URL.

        Note that we only do POSIX-style expansion here in order to maintain
        consistency across platforms.

        >>> os.environ["MY_B_VAR"] = "value"
        >>> Channel.from_string(
        ...     "https://host.com/$MY_B_VAR/channel"
        ... ).env_replaced_url()
        'https://host.com/value/channel'
        >>> Channel.from_string(
        ...     "https://host.com/${MY_B_VAR}/channel"
        ... ).env_replaced_url()
        'https://host.com/value/channel'
        >>> del os.environ["MY_B_VAR"]
        """
        return expandvars(self.url)

    def conda_token_replaced_url(self) -> str:
        """Emulate conda's token replacement in the output URL.

        This is used to recognize URLs contained in explicit lockfiles created by
        conda. In these lockfiles, the token is censored with <TOKEN>.

        >>> Channel.from_string(
        ...     "https://host.com/t/asdfjkl/channel"
        ... ).conda_token_replaced_url()
        'https://host.com/t/<TOKEN>/channel'
        """
        expanded_url = self.env_replaced_url()
        return mask_anaconda_token(expanded_url)

    def mamba_v1_token_replaced_url(self) -> str:
        """Emulate mamba's v1 token replacement in the output URL.

        This is used to recognize URLs contained in explicit lockfiles created by
        mamba or micromamba. In these lockfiles, the token is censored with *****.

        >>> Channel.from_string(
        ...     "https://host.com/t/asdfjkl/channel"
        ... ).mamba_v1_token_replaced_url()
        'https://host.com/t/*****/channel'
        """
        expanded_url = self.env_replaced_url()
        _, token = split_anaconda_token(expanded_url)
        return expanded_url.replace(token, "*****", 1) if token else expanded_url

    def mamba_v2_token_replaced_url(self) -> str:
        """Emulate mamba's v2 token replacement in the output URL.

        This is used to recognize URLs contained in explicit lockfiles created by
        mamba or micromamba. In these lockfiles, the token is censored with **********.

        >>> Channel.from_string(
        ...     "https://host.com/t/asdfjkl/channel"
        ... ).mamba_v2_token_replaced_url()
        'https://host.com/t/**********/channel'
        """
        expanded_url = self.env_replaced_url()
        _, token = split_anaconda_token(expanded_url)
        return expanded_url.replace(token, "**********", 1) if token else expanded_url

    def __repr_args__(self) -> list[tuple[str, Any]]:
        """Hide falsy values from repr.

        # Note how used_env_vars is not shown:
        >>> Channel.from_string("conda-forge")
        Channel(url='conda-forge')

        # It's only shown when it's non-empty:
        >>> Channel.from_string(
        ...     "https://host.com/t/$MY_REPO_TOKEN/channel"
        ... )  # doctest: +NORMALIZE_WHITESPACE
        Channel(url='https://host.com/t/${MY_REPO_TOKEN}/channel',
            used_env_vars=('MY_REPO_TOKEN',))
        """
        return [(key, value) for key, value in self.__dict__.items() if value]


def _detect_used_env_var(
    value: str, preferred_env_var_suffix: list[str]
) -> Optional[str]:
    """Detect if the string exactly matches any current environment variable.

    Preference is given to variables that end in the provided suffixes.
    """
    if value.startswith("$"):
        return value.lstrip("$").strip("{}")

    for suffix in [*preferred_env_var_suffix, ""]:
        candidates = {v: k for k, v in os.environ.items() if k.upper().endswith(suffix)}
        # Try first with a simple match
        if key := candidates.get(value):
            return key
        # Try with unquote
        if key := candidates.get(unquote(value)):
            return key
    return None


def _conda_url_from_string(url: str) -> _CondaUrl:
    """Normalize URL by using environment variables."""
    res = urlparse(url)

    def make_netloc(
        username: Optional[str], password: Optional[str], host: str, port: Optional[int]
    ) -> str:
        host_info = f"{host}:{port}" if port else host
        if not username:
            return host_info

        user_info = f"{username}:{password}" if password else username
        return f"{user_info}@{host_info}"

    user_env_var: Optional[str] = None
    password_env_var: Optional[str] = None
    token_env_var: Optional[str] = None

    def get_or_raise(val: Optional[str]) -> str:
        if val is None:
            raise ValueError("Expected to be non-null")
        return val

    res_replaced = copy.copy(res)

    if res.username:
        user_env_var = _detect_used_env_var(res.username, ["USERNAME", "USER"])
        if user_env_var:
            res_replaced = res_replaced._replace(
                netloc=make_netloc(
                    username=f"${{{user_env_var}}}",
                    password=res_replaced.password,
                    host=get_or_raise(res_replaced.hostname),
                    port=res_replaced.port,
                )
            )

    if res.password:
        password_env_var = _detect_used_env_var(
            res.password, ["PASSWORD", "PASS", "TOKEN", "KEY"]
        )
        if password_env_var:
            res_replaced = res_replaced._replace(
                netloc=make_netloc(
                    username=res_replaced.username,
                    password=f"${{{password_env_var}}}",
                    host=get_or_raise(res_replaced.hostname),
                    port=res_replaced.port,
                )
            )

    _token_match = token_pattern.search(res.path)
    token = _token_match.groups()[1][3:] if _token_match else None
    if token:
        token_env_var = _detect_used_env_var(
            token, ["TOKEN", "CRED", "PASSWORD", "PASS", "KEY"]
        )
        if not token_env_var:
            # Maybe we should raise here if we have mismatched env vars
            logger.warning("Token URL detected without env var")
        else:
            new_path = token_pattern.sub(
                rf"\1/t/${{{token_env_var}}}\3", res_replaced.path
            )
            res_replaced = res_replaced._replace(path=new_path)

    return _CondaUrl(
        raw_url=url,
        env_var_url=urlunparse(res_replaced),
        user=res.username,
        user_env_var=user_env_var,
        password=res.password,
        password_env_var=password_env_var,
        token=token,
        token_env_var=token_env_var,
    )


def _channel_from_conda_url(conda_url: _CondaUrl) -> Channel:
    conda_url = conda_url
    env_vars_maybe_with_none = {
        conda_url.user_env_var,
        conda_url.token_env_var,
        conda_url.password_env_var,
    }
    env_vars = {v for v in env_vars_maybe_with_none if v is not None}
    return Channel(
        url=conda_url.env_var_url,
        used_env_vars=tuple(sorted(env_vars)),
    )


def normalize_url_with_placeholders(url: str, channels: list[Channel]) -> str:
    """Normalize URLs from lockfiles to have env var placeholders.

    The env vars are associated with channel objects, so we do the replacement for
    each channel object individually.

    Conda lockfiles with tokens are censored to look like this:
    >>> conda_url = (
    ...     "http://localhost:32817/t/<TOKEN>/get/proxy-channel"
    ...     "/linux-64/zlib-1.3.1-hb9d3cd8_2.conda#c9f075ab2f33b3bbee9e62d4ad0a6cd8"
    ... )

    Mamba v1 lockfiles are censored with ***** instead of <TOKEN>
    >>> mamba_v1_url = conda_url.replace("<TOKEN>", "*****")

    Mamba v2 lockfiles are censored with ********** instead of <TOKEN>
    >>> mamba_v2_url = conda_url.replace("<TOKEN>", "**********")

    Create a channel with a token stored in an env var
    >>> os.environ["MY_C_REPO_TOKEN"] = "some-token"
    >>> channel_url = "http://localhost:32817/t/some-token/get/proxy-channel"
    >>> channel = Channel.from_string(channel_url)
    >>> channel  # doctest: +NORMALIZE_WHITESPACE
    Channel(url='http://localhost:32817/t/${MY_C_REPO_TOKEN}/get/proxy-channel',
        used_env_vars=('MY_C_REPO_TOKEN',))

    The normalized URL should have the token replaced with the env var placeholder
    >>> expected_normalized_url = conda_url.replace("<TOKEN>", "${MY_C_REPO_TOKEN}")

    Check that the normalized URL is correct for both conda and mamba censorings
    >>> normalized_conda_url = normalize_url_with_placeholders(conda_url, [channel])
    >>> normalized_mamba_v1_url = normalize_url_with_placeholders(mamba_v1_url, [channel])
    >>> normalized_mamba_v2_url = normalize_url_with_placeholders(mamba_v2_url, [channel])
    >>> assert normalized_conda_url == expected_normalized_url, normalized_conda_url
    >>> assert normalized_mamba_v1_url == expected_normalized_url, normalized_mamba_v1_url
    >>> assert normalized_mamba_v2_url == expected_normalized_url, normalized_mamba_v2_url

    Normalization should also work similarly for basic auth
    >>> os.environ["MY_C_USERNAME"] = "user"
    >>> os.environ["MY_C_PASSWORD"] = "pass"
    >>> channel_url = "http://user:pass@localhost:32817/channel"
    >>> channel = Channel.from_string(channel_url)
    >>> channel  # doctest: +NORMALIZE_WHITESPACE
    Channel(url='http://${MY_C_USERNAME}:${MY_C_PASSWORD}@localhost:32817/channel',
        used_env_vars=('MY_C_PASSWORD', 'MY_C_USERNAME'))
    >>> normalize_url_with_placeholders(channel_url, channels=[channel])
    'http://${MY_C_USERNAME}:${MY_C_PASSWORD}@localhost:32817/channel'

    Clean up
    >>> del os.environ["MY_C_REPO_TOKEN"]
    >>> del os.environ["MY_C_USERNAME"]
    >>> del os.environ["MY_C_PASSWORD"]
    """
    for channel in channels:
        candidate1 = channel.conda_token_replaced_url()
        if url.startswith(candidate1):
            url = url.replace(candidate1, channel.url, 1)

        candidate2 = channel.mamba_v1_token_replaced_url()
        if url.startswith(candidate2):
            url = url.replace(candidate2, channel.url, 1)

        candidate3 = channel.mamba_v2_token_replaced_url()
        if url.startswith(candidate3):
            url = url.replace(candidate3, channel.url, 1)

        candidate4 = channel.env_replaced_url()
        if url.startswith(candidate4):
            url = url.replace(candidate4, channel.url, 1)
    return url



from __future__ import annotations

from conda_lock._vendor.poetry.core.exceptions import PoetryCoreError
from tomlkit.exceptions import TOMLKitError


class TOMLError(TOMLKitError, PoetryCoreError):
    pass



import json
import logging
import os
import pathlib
import shlex
import shutil
import subprocess
import sys
import tempfile
import time

from collections.abc import Iterable, Iterator, MutableSequence, Sequence
from contextlib import contextmanager
from textwrap import dedent
from typing import (
    Any,
    Literal,
    Optional,
)
from urllib.parse import urlsplit, urlunsplit

from conda_lock.interfaces.vendored_conda import MatchSpec
from conda_lock.invoke_conda import (
    PathLike,
    _get_conda_flags,
    conda_env_override,
    conda_pkgs_dir,
    is_micromamba,
)
from conda_lock.lockfile import apply_categories
from conda_lock.lockfile.v2prelim.models import HashModel, LockedDependency
from conda_lock.models.channel import Channel, normalize_url_with_placeholders
from conda_lock.models.dry_run_install import DryRunInstall, FetchAction, LinkAction
from conda_lock.models.lock_spec import Dependency, VersionedDependency


logger = logging.getLogger(__name__)


def _to_match_spec(
    conda_dep_name: str,
    conda_version: Optional[str],
    build: Optional[str],
    conda_channel: Optional[str],
) -> str:
    kwargs = dict(name=conda_dep_name)
    if conda_version:
        kwargs["version"] = conda_version
    if build:
        kwargs["build"] = build
        if "version" not in kwargs:
            kwargs["version"] = "*"
    if conda_channel:
        kwargs["channel"] = conda_channel

    ms = MatchSpec(**kwargs)  # pyright: ignore[reportArgumentType]
    # Since MatchSpec doesn't round trip to the cli well
    if conda_channel:
        # this will return "channel_name::package_name"
        return str(ms)
    else:
        # this will return only "package_name" even if there's a channel in the kwargs
        return ms.conda_build_form()


def extract_json_object(proc_stdout: str) -> str:
    try:
        return proc_stdout[proc_stdout.index("{") : proc_stdout.rindex("}") + 1]
    except ValueError:
        return proc_stdout


def solve_conda(
    conda: PathLike,
    specs: dict[str, Dependency],
    locked: dict[str, LockedDependency],
    update: list[str],
    platform: str,
    channels: list[Channel],
    mapping_url: str,
) -> dict[str, LockedDependency]:
    """
    Solve (or update a previous solution of) conda specs for the given platform

    Parameters
    ----------
    conda :
        Path to conda, mamba, or micromamba
    specs :
        Conda package specifications
    locked :
        Previous solution for the given platform (conda packages only)
    update :
        Named of packages to update to the latest version compatible with specs
    platform :
        Target platform
    channels :
        Channels to query

    """

    conda_specs = [
        _to_match_spec(dep.name, dep.version, dep.build, dep.conda_channel)
        for dep in specs.values()
        if isinstance(dep, VersionedDependency) and dep.manager == "conda"
    ]
    conda_locked = {dep.name: dep for dep in locked.values() if dep.manager == "conda"}
    to_update = set(update).intersection(conda_locked)

    if to_update:
        dry_run_install = update_specs_for_arch(
            conda=conda,
            platform=platform,
            channels=channels,
            specs=conda_specs,
            locked=conda_locked,
            update=list(to_update),
        )
    else:
        dry_run_install = solve_specs_for_arch(
            conda=conda,
            platform=platform,
            channels=channels,
            specs=conda_specs,
        )
    logging.debug("dry_run_install:\n%s", dry_run_install)

    # extract dependencies from package plan
    planned = {}
    for action in dry_run_install["actions"]["FETCH"]:
        dependencies = {}
        for dep in action.get("depends") or []:
            matchspec = MatchSpec(dep)  # pyright: ignore[reportArgumentType]
            name = matchspec.name
            version = (
                matchspec.version.spec_str if matchspec.version is not None else ""
            )
            dependencies[name] = version

        locked_dependency = LockedDependency(
            name=action["name"],
            version=action["version"],
            manager="conda",
            platform=platform,
            dependencies=dependencies,
            # TODO: Normalize URL here and inject env vars
            url=normalize_url_with_placeholders(action["url"], channels=channels),
            # NB: virtual packages may have no hash
            hash=HashModel(
                md5=action["md5"] if "md5" in action else "",
                sha256=action.get("sha256"),
            ),
        )
        planned[action["name"]] = locked_dependency

    # propagate categories from explicit to transitive dependencies
    apply_categories(
        requested={k: v for k, v in specs.items() if v.manager == "conda"},
        planned=planned,
        mapping_url=mapping_url,
    )

    return planned


def _get_repodata_record(
    pkgs_dirs: list[pathlib.Path], dist_name: str
) -> Optional[FetchAction]:
    """Get the repodata_record.json of a given distribution from the package cache.

    On rare occasion during the CI tests, conda fails to find a package in the
    package cache, perhaps because the package is still being processed? Waiting for
    0.1 seconds seems to solve the issue. Here we allow for a full second to elapse
    before giving up.
    """
    NUM_RETRIES = 10
    for retry in range(1, NUM_RETRIES + 1):
        for pkgs_dir in pkgs_dirs:
            record = pkgs_dir / dist_name / "info" / "repodata_record.json"
            if record.exists():
                with open(record) as f:
                    repodata: FetchAction = json.load(f)
                return repodata
        logger.warning(
            f"Failed to find repodata_record.json for {dist_name}. "
            f"Retrying in 0.1 seconds ({retry}/{NUM_RETRIES})"
        )
        time.sleep(0.1)
    logger.warning(f"Failed to find repodata_record.json for {dist_name}. Giving up.")
    return None


def _get_pkgs_dirs(
    *,
    conda: PathLike,
    platform: str,
    method: Optional[Literal["config", "info"]] = None,
) -> list[pathlib.Path]:
    """Extract the package cache directories from the conda configuration."""
    if method is None:
        method = "config" if is_micromamba(conda) else "info"
    if method == "config":
        # 'package cache' was added to 'micromamba info' in v1.4.6.
        args = [str(conda), "config", "--json", "list", "pkgs_dirs"]
    elif method == "info":
        args = [str(conda), "info", "--json"]
    env = conda_env_override(platform)
    output = subprocess.check_output(args, env=env).decode()
    json_object_str = extract_json_object(output)
    json_object: dict[str, Any] = json.loads(json_object_str)
    pkgs_dirs_list: list[str]
    if "pkgs_dirs" in json_object:
        pkgs_dirs_list = json_object["pkgs_dirs"]
    elif "package cache" in json_object:
        pkgs_dirs_list = json_object["package cache"]
    else:
        raise ValueError(
            f"Unable to extract pkgs_dirs from {json_object}. "
            "Please report this issue to the conda-lock developers."
        )
    pkgs_dirs = [pathlib.Path(d) for d in pkgs_dirs_list]
    return pkgs_dirs


def _reconstruct_fetch_actions(
    conda: PathLike, platform: str, dry_run_install: DryRunInstall
) -> DryRunInstall:
    """
    Conda may choose to link a previously downloaded distribution from pkgs_dirs rather
    than downloading a fresh one. Find the repodata record in existing distributions
    that have only a LINK action, and use it to synthesize a corresponding FETCH action
    with the metadata we need to extract for the package plan.
    """
    if "LINK" not in dry_run_install["actions"]:
        dry_run_install["actions"]["LINK"] = []
    if "FETCH" not in dry_run_install["actions"]:
        dry_run_install["actions"]["FETCH"] = []

    link_actions = {p["name"]: p for p in dry_run_install["actions"]["LINK"]}
    fetch_actions = {p["name"]: p for p in dry_run_install["actions"]["FETCH"]}
    link_only_names = set(link_actions.keys()).difference(fetch_actions.keys())
    if link_only_names:
        pkgs_dirs = _get_pkgs_dirs(conda=conda, platform=platform)
    else:
        pkgs_dirs = []

    for link_pkg_name in link_only_names:
        link_action = link_actions[link_pkg_name]
        if "dist_name" in link_action:
            dist_name = link_action["dist_name"]
        elif "fn" in link_action:
            dist_name = str(link_action["fn"])
            if dist_name.endswith(".tar.bz2"):
                dist_name = dist_name[:-8]
            elif dist_name.endswith(".conda"):
                dist_name = dist_name[:-6]
            else:
                raise ValueError(f"Unknown filename format: {dist_name}")
        else:
            raise ValueError(f"Unable to extract the dist_name from {link_action}.")
        repodata = _get_repodata_record(pkgs_dirs, dist_name)
        if repodata is None:
            raise FileNotFoundError(
                f"Distribution '{dist_name}' not found in pkgs_dirs {pkgs_dirs}"
            )
        dry_run_install["actions"]["FETCH"].append(repodata)
    return dry_run_install


def solve_specs_for_arch(
    conda: PathLike,
    channels: Sequence[Channel],
    specs: list[str],
    platform: str,
) -> DryRunInstall:
    """
    Solve conda specifications for the given platform

    Parameters
    ----------
    conda :
        Path to conda, mamba, or micromamba
    channels :
        Channels to query
    specs :
        Conda package specifications
    platform :
        Target platform

    """
    args: MutableSequence[str] = [
        str(conda),
        "create",
        "--prefix",
        os.path.join(conda_pkgs_dir(), "prefix"),
        "--dry-run",
        "--json",
    ]
    args.extend(_get_conda_flags(channels=channels, platform=platform))
    args.extend(specs)
    logger.info("%s using specs %s", platform, specs)
    logger.debug(f"Running command {shlex.join(args)}")
    proc = subprocess.run(  # noqa: UP022  # Poetry monkeypatch breaks capture_output
        [str(arg) for arg in args],
        env=conda_env_override(platform),
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
        encoding="utf8",
    )

    def print_proc(proc: subprocess.CompletedProcess) -> None:
        print(f"    Command: {proc.args}", file=sys.stderr)
        if proc.stdout:
            print(f"    STDOUT:\n{proc.stdout}", file=sys.stderr)
        if proc.stderr:
            print(f"    STDERR:\n{proc.stderr}", file=sys.stderr)

    try:
        proc.check_returncode()
    except subprocess.CalledProcessError:
        try:
            err_json = json.loads(proc.stdout)
            try:
                message = err_json["message"]
            except KeyError:
                print("Message key not found in json! returning the full json text")
                message = err_json
        except json.JSONDecodeError as e:
            print(f"Failed to parse json, {e}", file=sys.stderr)
            message = proc.stdout

        print(
            f"Could not lock the environment for platform {platform}", file=sys.stderr
        )
        if message:
            print(message, file=sys.stderr)
        print_proc(proc)

        raise

    try:
        dryrun_install: DryRunInstall = json.loads(extract_json_object(proc.stdout))
        return _reconstruct_fetch_actions(conda, platform, dryrun_install)
    except json.JSONDecodeError as e:
        raise RuntimeError(f"Failed to parse json: '{proc.stdout}'") from e


def _get_installed_conda_packages(
    conda: PathLike,
    platform: str,
    prefix: str,
) -> dict[str, LinkAction]:
    """
    Get the installed conda packages for the given prefix.

    Try to get installed packages, first with --no-pip flag, then without if that fails.
    The --no-pip flag was added in Conda v2.1.0 (2013), but for mamba/micromamba only in
    v2.0.7 (March 2025).
    """
    try:
        output = subprocess.check_output(
            [str(conda), "list", "--no-pip", "-p", prefix, "--json"],
            env=conda_env_override(platform),
            stderr=subprocess.STDOUT,
        )
    except subprocess.CalledProcessError as e:
        err_output = (
            e.output.decode("utf-8") if isinstance(e.output, bytes) else e.output
        )
        if "The following argument was not expected: --no-pip" in err_output:
            logger.warning(
                f"The '--no-pip' flag is not supported by {conda}. Please consider upgrading."
            )
            # Retry without --no-pip
            output = subprocess.check_output(
                [str(conda), "list", "-p", prefix, "--json"],
                env=conda_env_override(platform),
                stderr=subprocess.STDOUT,
            )
        else:
            # Re-raise if it's a different error.
            raise
    decoded_output = output.decode("utf-8")
    installed: dict[str, LinkAction] = {
        entry["name"]: entry for entry in json.loads(decoded_output)
    }
    return installed


def update_specs_for_arch(
    conda: PathLike,
    specs: list[str],
    locked: dict[str, LockedDependency],
    update: list[str],
    platform: str,
    channels: Sequence[Channel],
) -> DryRunInstall:
    """
    Update a previous solution for the given platform

    Parameters
    ----------
    conda :
        Path to conda, mamba, or micromamba
    specs :
        Conda package specifications
    locked :
        Previous solution for the given platform (conda packages only)
    update :
        Named of packages to update to the latest version compatible with specs
    platform :
        Target platform
    channels :
        Channels to query

    """

    with fake_conda_environment(locked.values(), platform=platform) as prefix:
        installed = _get_installed_conda_packages(conda, platform, prefix)
        spec_for_name = {MatchSpec(v).name: v for v in specs}  # pyright: ignore
        to_update = [
            spec_for_name[name] for name in set(installed).intersection(update)
        ]
        if to_update:
            # NB: [micro]mamba and mainline conda have different semantics for `install` and `update`
            # - conda:
            #   * update -> apply all nonmajor updates unconditionally (unless pinned)
            #   * install -> install or update target to latest version compatible with constraint
            # - micromamba:
            #   * update -> update target to latest version compatible with constraint
            #   * install -> update target if current version incompatible with constraint, otherwise _do nothing_
            # - mamba:
            #   * update -> apply all nonmajor updates unconditionally (unless pinned)
            #   * install -> update target if current version incompatible with constraint, otherwise _do nothing_
            # Our `update` should always update the target to the latest version compatible with the constraint,
            # while updating as few other packages as possible. With mamba this can only be done with pinning.
            if pathlib.Path(conda).name.startswith("mamba"):
                # pin non-updated packages to prevent _any_ movement
                pinned_filename = pathlib.Path(prefix) / "conda-meta" / "pinned"
                assert not pinned_filename.exists()
                with open(pinned_filename, "w") as pinned:
                    for name in set(installed.keys()).difference(update):
                        pinned.write(f"{name} =={installed[name]['version']}\n")
                args = [
                    str(conda),
                    "update",
                    *_get_conda_flags(channels=channels, platform=platform),
                ]
                print(
                    "Warning: mamba cannot update single packages without resorting to pinning. "
                    "If the update fails to solve, try with conda or micromamba instead.",
                    file=sys.stderr,
                )
            else:
                args = [
                    str(conda),
                    "update" if is_micromamba(conda) else "install",
                    *_get_conda_flags(channels=channels, platform=platform),
                ]
            cmd = [
                str(arg)
                for arg in [*args, "-p", prefix, "--json", "--dry-run", *to_update]
            ]
            logger.debug(f"Running command {shlex.join(cmd)}")
            proc = subprocess.run(  # noqa: UP022  # Poetry monkeypatch breaks capture_output
                cmd,
                env=conda_env_override(platform),
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE,
                encoding="utf8",
            )

            try:
                proc.check_returncode()
            except subprocess.CalledProcessError as exc:
                err_json = json.loads(proc.stdout)
                raise RuntimeError(
                    f"Could not lock the environment for platform {platform}: {err_json.get('message')}"
                ) from exc

            dryrun_install: DryRunInstall = json.loads(extract_json_object(proc.stdout))
        else:
            dryrun_install = {"actions": {"LINK": [], "FETCH": []}}

        if "actions" not in dryrun_install:
            dryrun_install["actions"] = {"LINK": [], "FETCH": []}

        updated = {entry["name"]: entry for entry in dryrun_install["actions"]["LINK"]}
        for package in set(installed).difference(updated):
            # This is the case where the package is unchanged.
            # First create a FETCH action based on the original lockfile entry.
            original_lockfile_entry = locked[package]
            original_fetch_action = original_lockfile_entry.to_fetch_action()
            dryrun_install["actions"]["FETCH"].append(original_fetch_action)

            # Then create a LINK action to indicate that the package is installed.
            installed_link_action = installed[package]
            dryrun_install["actions"]["LINK"].append(installed_link_action)

        reconstructed = _reconstruct_fetch_actions(conda, platform, dryrun_install)
        return reconstructed


@contextmanager
def fake_conda_environment(
    locked: Iterable[LockedDependency], platform: str
) -> Iterator[str]:
    """
    Create a fake conda prefix containing metadata corresponding to the provided dependencies

    Parameters
    ----------
    locked :
        Previous solution
    platform :
        Target platform

    """
    with tempfile.TemporaryDirectory() as prefix:
        conda_meta = pathlib.Path(prefix) / "conda-meta"
        conda_meta.mkdir()
        (conda_meta / "history").touch()
        for dep in (
            dep for dep in locked if dep.manager == "conda" and dep.platform == platform
        ):
            url = urlsplit(dep.url)
            path = pathlib.PurePosixPath(url.path)
            channel = urlunsplit(
                (url.scheme, url.hostname, str(path.parent), None, None)
            )
            truncated_path = path
            while truncated_path.suffix in {".tar", ".bz2", ".gz", ".conda"}:
                truncated_path = truncated_path.with_suffix("")
            build = truncated_path.name.split("-")[-1]
            try:
                build_number = int(build.split("_")[-1])
            except ValueError:
                build_number = 0
            entry = {
                "name": dep.name,
                "channel": channel,
                "url": dep.url,
                "md5": dep.hash.md5,
                "build": build,
                "build_number": build_number,
                "version": dep.version,
                "subdir": path.parent.name,
                "fn": path.name,
                "depends": [f"{k} {v}".strip() for k, v in dep.dependencies.items()],
            }
            # mamba requires these to be stringlike so null are not allowed here
            if dep.hash.sha256 is not None:
                entry["sha256"] = dep.hash.sha256

            with open(conda_meta / (truncated_path.name + ".json"), "w") as f:
                json.dump(entry, f, indent=2)
            make_fake_python_binary(prefix)
        yield prefix


def make_fake_python_binary(prefix: str) -> None:
    """Create a fake python binary in the given prefix.

    Our fake Conda environment contains metadata indicating that `python`
    is installed in the prefix, however no packages are installed.

    This is intended to prevent failure of `PrefixData.load_site_packages`
    which was introduced in libmamba v2. That function invokes the command
    `python -q -m pip inspect --local` to check for installed pip packages,
    where the `python` binary is the one in the conda prefix. Our fake
    prefix only contains the package metadata records, not the actual
    packages or binaries. At this stage for conda-lock, we are only
    interested in the conda packages, so we spoof the `python` binary
    to return an empty stdout so that things proceed without error.
    """
    # Write the fake Python script to a file
    fake_python_script = pathlib.Path(prefix) / "fake_python_script.py"
    fake_python_script.write_text(
        dedent(
            """\
            import sys
            import shlex

            cmd = shlex.join(sys.argv)

            stderr_message = f'''\
            This is a fake python binary generated by conda-lock.

            It prevents libmamba from failing when it tries to check for installed \
            pip packages.

            For more details, see the docstring for `make_fake_python_binary`.

            This was called as:
                {cmd}
            '''

            print(stderr_message, file=sys.stderr, flush=True, end='')

            if "-m pip" in cmd:
                # Simulate an empty `pip inspect` output
                print('{}', flush=True)
            else:
                raise RuntimeError("Expected to invoke pip module with `-m pip`.")
            """
        )
    )

    if sys.platform == "win32":
        # On Windows, copy sys.executable to prefix/Scripts/python.exe
        fake_python_binary = pathlib.Path(prefix) / "Scripts" / "python.exe"
        fake_python_binary.parent.mkdir(parents=True, exist_ok=True)
        shutil.copyfile(sys.executable, fake_python_binary)

        # Adjust the environment to ensure our fake script is executed
        # Create a wrapper batch file that sets PYTHONPATH
        wrapper_batch = pathlib.Path(prefix) / "Scripts" / "python.bat"
        wrapper_batch_content = dedent(f"""\
            @echo off
            set PYTHONPATH={prefix};%PYTHONPATH%
            "{fake_python_binary}" %*
        """)
        wrapper_batch.write_text(wrapper_batch_content)
    else:
        # On Unix-like systems, create a shell script that calls the script
        fake_python_binary = pathlib.Path(prefix) / "bin" / "python"
        fake_python_binary.parent.mkdir(parents=True, exist_ok=True)
        shell_script_content = dedent(f"""\
            #!/usr/bin/env sh
            "{sys.executable}" "{fake_python_script}" "$@"
        """)
        fake_python_binary.write_text(shell_script_content)
        fake_python_binary.chmod(0o755)



from collections import defaultdict
from typing import Any, NamedTuple, Optional, Union

from tomlkit import TOMLDocument, comment, document, inline_table, item, table
from tomlkit.items import InlineTable, Table

from conda_lock._export_lock_spec_compute_platform_indep import (
    EditableDependency,
    unify_platform_independent_deps,
)
from conda_lock.common import warn
from conda_lock.models.lock_spec import (
    Dependency,
    LockSpecification,
    URLDependency,
    VCSDependency,
    VersionedDependency,
)


class TomlTableKey(NamedTuple):
    """Represents a key in a pixi.toml table.

    It can be rendered into a TOML header using `toml_header_sequence`.

    >>> toml_header_sequence(
    ...     TomlTableKey(category="dev", platform="linux-64", manager="pip")
    ... )
    ['feature', 'dev', 'target', 'linux-64', 'pypi-dependencies']
    """

    category: str
    platform: Optional[str]
    manager: str


def render_pixi_toml(
    *,
    lock_spec: LockSpecification,
    project_name: Optional[str] = None,
    with_cuda: Optional[str] = None,
    editables: Optional[list[EditableDependency]] = None,
) -> TOMLDocument:
    """Render a pixi.toml from a LockSpecification as a tomlkit TOMLDocument."""
    pixi_toml = document()
    for line in (
        "This file was generated by conda-lock for the pixi environment manager.",
        "For more information, see <https://github.com/conda/conda-lock> "
        "and <https://pixi.sh>.",
        "Source files:",
        *(f"- {src_file.as_posix()}" for src_file in lock_spec.sources),
    ):
        pixi_toml.add(comment(line))

    if project_name is None:
        project_name = "project-name-placeholder"
    all_platforms = sorted(lock_spec.dependencies.keys())
    all_categories: set[str] = set()
    for platform in all_platforms:
        for dep in lock_spec.dependencies[platform]:
            all_categories.add(dep.category)
    if {"main", "default"} <= all_categories:
        # The default "main" category of conda-lock will be rendered as
        # the "default" feature in pixi. If both are defined, then this
        # is a conflict.
        raise ValueError("Cannot have both 'main' and 'default' as categories/extras")

    pixi_toml.add(
        "project",
        item(
            dict(
                name=project_name,
                platforms=list(all_platforms),
                channels=[channel.url for channel in lock_spec.channels],
            )
        ),
    )
    if len(lock_spec.channels) == 0:
        warn(
            "No channels defined in the lock file! Consider defining channels, e.g. "
            "via the command line by adding `--channel conda-forge`."
        )
    for channel in lock_spec.channels:
        if channel.used_env_vars:
            warn(
                f"Channel {channel.url} uses environment variables, which will "
                "be dropped in the pixi.toml."
            )

    # The dependency tables
    arranged_deps = arrange_for_toml(lock_spec, editables=editables)
    for key, deps_by_name in arranged_deps.items():
        header_sequence: list[str] = toml_header_sequence(key)

        # Keys are package names, values are version numbers or matchspecs.
        inner_dict = {
            name: toml_dependency_value(dep) for name, dep in deps_by_name.items()
        }

        # Using the nested sequence of headers, walk down the tree of tomlkit tables
        # towards the leaf where the inner_dict should be inserted.
        node: Union[TOMLDocument, Table] = pixi_toml
        for header in header_sequence:
            if header not in node:
                node.add(header, table())
            next_node = node[header]
            assert isinstance(next_node, Table)
            node = next_node
        # Now `node` is the leaf table where the inner table should be inserted.
        node.update(inner_dict)

    # The environments table
    if len(all_categories) > 1:
        pixi_toml.add("environments", toml_environments_table(all_categories))

    # The system requirements table
    if with_cuda:
        pixi_toml.add("system-requirements", item(dict(cuda=with_cuda)))

    return pixi_toml


def toml_dependency_value(
    dep: Union[Dependency, EditableDependency],
) -> Union[str, InlineTable]:
    """Render a conda-lock Dependency as a pixi.toml line as VersionSpec or matchspec.

    The result is suitable for the values used in the `dependencies` or
    `pypi-dependencies` tables of a pixi TOML file.

    >>> toml_dependency_value(VersionedDependency(name="numpy", version="2.1.1"))
    '2.1.1'

    >>> toml_dependency_value(VersionedDependency(name="numpy", version=""))
    '*'

    >>> toml_dependency_value(
    ...     VersionedDependency(
    ...         name="numpy",
    ...         version="2.1.1",
    ...         conda_channel="conda-forge",
    ...         build="py313h4bf6692_0"
    ...     )
    ... )
    {'version': '2.1.1', 'build': 'py313h4bf6692_0', 'channel': 'conda-forge'}

    >>> toml_dependency_value(
    ...     VersionedDependency(
    ...         name="xarray",
    ...         version="",
    ...         extras=["io", "parallel"],
    ...         manager="pip",
    ...     )
    ... )
    {'version': '*', 'extras': ['io', 'parallel']}
    """
    matchspec: dict[str, Any] = {}
    if isinstance(dep, VersionedDependency):
        matchspec["version"] = dep.version or "*"
        if dep.manager == "pip" and matchspec["version"][0].isdigit():
            matchspec["version"] = f"=={matchspec['version']}"
        if dep.build is not None:
            matchspec["build"] = dep.build
        if dep.hash is not None:
            raise NotImplementedError(f"Hash not yet supported in {dep}")
        if dep.conda_channel is not None:
            matchspec["channel"] = dep.conda_channel
        if dep.extras:
            if dep.extras and dep.manager == "conda":
                warn(f"Extras not supported in Conda dep {dep}")
            else:
                matchspec["extras"] = dep.extras
        if len(matchspec) == 1:
            # Use the simpler VersionSpec format if there's only a version.
            return matchspec["version"]
        else:
            return _dict_to_inline_table(matchspec)
    elif isinstance(dep, URLDependency):
        raise NotImplementedError(f"URL not yet supported in {dep}")
    elif isinstance(dep, VCSDependency):
        raise NotImplementedError(f"VCS not yet supported in {dep}")
    elif isinstance(dep, EditableDependency):
        return _dict_to_inline_table(dict(path=dep.path, editable=True))
    else:
        raise ValueError(f"Unknown dependency type {dep}")


def _dict_to_inline_table(d: dict[str, Any]) -> InlineTable:
    """Convert a dictionary to a TOML inline table."""
    table = inline_table()
    table.update(d)
    return table


def arrange_for_toml(
    lock_spec: LockSpecification,
    *,
    editables: Optional[list[EditableDependency]] = None,
) -> dict[TomlTableKey, dict[str, Union[Dependency, EditableDependency]]]:
    """Arrange dependencies into a structured dictionary for TOML generation."""
    unified_deps = unify_platform_independent_deps(
        lock_spec.dependencies, editables=editables
    )

    # Stick all the dependencies into the correct TOML table
    unsorted_result: dict[
        TomlTableKey, dict[str, Union[Dependency, EditableDependency]]
    ] = defaultdict(dict)
    for dep_key, dep in unified_deps.items():
        toml_key = TomlTableKey(
            category=dep_key.category,
            platform=dep_key.platform,
            manager=dep_key.manager,
        )
        if dep_key.name in unsorted_result[toml_key]:
            # This should never happen. Keys must be unique within a TOML table.
            # Moreover, the `unify_platform_independent_deps` function should have
            # already ensured that there are no duplicate keys.
            preexisting_dep = unsorted_result[toml_key][dep_key.name]
            raise RuntimeError(
                f"Duplicate key {dep_key} for {dep} and {preexisting_dep}"
            )
        unsorted_result[toml_key][dep_key.name] = dep

    # Alphabetize the dependencies within each table
    alphabetized_result = {
        toml_key: dict(sorted(deps_by_name.items()))
        for toml_key, deps_by_name in unsorted_result.items()
    }

    # Sort the tables themselves
    sorted_result = dict(sorted(alphabetized_result.items(), key=toml_ordering))

    return sorted_result


def toml_ordering(item: tuple[TomlTableKey, dict]) -> tuple[str, str, str]:
    """Make a sort key to properly order the dependency tables in the pixi.toml.

    The main category = default feature comes first. Then the other categories.

    Within each category, the platform-independent dependencies come first, followed by
    the platform-specific dependencies.

    Within each platform, we declare the conda dependencies first, followed by the pip
    dependencies.

    Within each table determined by the hierarchy of category, platform, and manager,
    the dependencies are sorted alphabetically by name. But the key here is just for
    sorting the tables that occur, not the dependencies within them.

    We define the ordering via a tuple of strings: (category, platform, manager).
    The main category and the platform-independent dependencies are represented by
    empty strings so that they come lexicographically first.

    >>> toml_ordering(
    ...     (TomlTableKey(category="main", platform=None, manager="conda"), {})
    ... )
    ('', '', 'conda')

    >>> toml_ordering(
    ...     (TomlTableKey(category="main", platform="linux-64", manager="conda"), {})
    ... )
    ('', 'linux-64', 'conda')

    >>> toml_ordering((TomlTableKey(category="dev", platform=None, manager="pip"), {}))
    ('dev', '', 'pip')
    """
    key = item[0]
    category = "" if key.category in ["main", "default"] else key.category
    platform = key.platform if key.platform is not None else ""
    # "conda" before "pip" is conveniently already lexicographical order.
    return category, platform, key.manager


def toml_header_sequence(key: TomlTableKey) -> list[str]:
    """Generates a TOML header based on the dependency type, platform, and manager.

    >>> toml_header_sequence(
    ...     TomlTableKey(category="main", platform=None, manager="conda")
    ... )
    ['dependencies']

    >>> toml_header_sequence(
    ...     TomlTableKey(category="main", platform="linux-64", manager="conda")
    ... )
    ['target', 'linux-64', 'dependencies']

    >>> toml_header_sequence(
    ...     TomlTableKey(category="main", platform=None, manager="pip")
    ... )
    ['pypi-dependencies']

    >>> toml_header_sequence(
    ...     TomlTableKey(category="main", platform="linux-64", manager="pip")
    ... )
    ['target', 'linux-64', 'pypi-dependencies']

    >>> toml_header_sequence(
    ...     TomlTableKey(category="dev", platform=None, manager="conda")
    ... )
    ['feature', 'dev', 'dependencies']

    >>> toml_header_sequence(
    ...     TomlTableKey(category="dev", platform="linux-64", manager="conda")
    ... )
    ['feature', 'dev', 'target', 'linux-64', 'dependencies']

    >>> toml_header_sequence(TomlTableKey(category="dev", platform=None, manager="pip"))
    ['feature', 'dev', 'pypi-dependencies']

    >>> toml_header_sequence(
    ...     TomlTableKey(category="dev", platform="linux-64", manager="pip")
    ... )
    ['feature', 'dev', 'target', 'linux-64', 'pypi-dependencies']
    """
    parts = []
    if key.category not in ["main", "default"]:
        parts.extend(["feature", key.category])
    if key.platform:
        parts.extend(["target", key.platform])
    parts.append("dependencies" if key.manager == "conda" else "pypi-dependencies")
    return parts


def toml_environments_table(all_categories: set[str]) -> Table:
    r"""Define the environments section of a pixi.toml file.

    >>> environments_table = toml_environments_table({"main", "dev", "docs"})
    >>> print(environments_table.as_string())
    # Redefine the default environment to include all categories.
    default = ["dev", "docs"]
    # Define a minimal environment with only the default feature.
    minimal = []
    # Create an environment for each feature.
    dev = ["dev"]
    docs = ["docs"]
    <BLANKLINE>
    """
    non_default_categories = sorted(all_categories - {"main", "default"})
    if len(non_default_categories) == 0:
        raise ValueError("Expected at least one non-default category")

    environments = table()
    environments.add(
        comment("Redefine the default environment to include all categories.")
    )
    environments.add("default", non_default_categories)

    MINIMAL_ENVIRONMENT_NAMES = ["minimal", "prod", "main"]
    minimal_category_name = next(
        (name for name in MINIMAL_ENVIRONMENT_NAMES if name not in all_categories), None
    )
    if minimal_category_name is None:
        warn(
            "Can't find a name for the 'minimal' environment since categories for '"
            + "', '".join(MINIMAL_ENVIRONMENT_NAMES)
            + "' are already defined. Skipping."
        )
    else:
        environments.add(
            comment("Define a minimal environment with only the default feature.")
        )
        environments.add(minimal_category_name, [])

    environments.add(comment("Create an environment for each feature."))
    for category in non_default_categories:
        environments.add(category, [category])
    return environments



from conda_lock._vendor.poetry.core.version.markers import (
    AnyMarker,
    BaseMarker,
    EmptyMarker,
    MarkerUnion,
    MultiMarker,
    SingleMarker,
    parse_marker,
)


__all__ = [
    "AnyMarker",
    "BaseMarker",
    "EmptyMarker",
    "MarkerUnion",
    "MultiMarker",
    "SingleMarker",
    "parse_marker",
]



import hashlib
import logging
import re

from datetime import datetime
from pathlib import Path
from typing import Optional

import requests

from filelock import FileLock, Timeout
from platformdirs import user_cache_path


logger = logging.getLogger(__name__)


CLEAR_CACHE_AFTER_SECONDS = 60 * 60 * 24 * 2  # 2 days
"""Cached files older than this will be deleted."""

DONT_CHECK_IF_NEWER_THAN_SECONDS = 60 * 5  # 5 minutes
"""If the cached file is newer than this, just use it without checking for updates."""

TIMESTAMP_RESOLUTION_SECONDS = 5
"""Some filesystems may have a limited resolution for file modification times.
We tolerate a small amount of timestamp inaccuracy that might lead to files having
a negative age.
"""


def uncached_download_file(url: str) -> bytes:
    """The simple equivalent to cached_download_file."""
    res = requests.get(url, headers={"User-Agent": "conda-lock"})
    res.raise_for_status()
    return res.content


def cached_download_file(
    url: str,
    *,
    cache_subdir_name: str,
    cache_root: Optional[Path] = None,
    max_age_seconds: float = CLEAR_CACHE_AFTER_SECONDS,
    dont_check_if_newer_than_seconds: float = DONT_CHECK_IF_NEWER_THAN_SECONDS,
) -> bytes:
    """Download a file and cache it in the user cache directory.

    If the file is already cached, return the cached contents.
    If the file is not cached, download it and cache the contents
    and the ETag.

    Protect against multiple processes downloading the same file.
    """
    if cache_root is None:
        cache_root = user_cache_path("conda-lock", appauthor=False)
    cache = cache_root / "cache" / cache_subdir_name
    cache.mkdir(parents=True, exist_ok=True)
    clear_old_files_from_cache(cache, max_age_seconds=max_age_seconds)

    destination = cache / cached_filename_for_url(url)
    destination_lock = destination.with_suffix(".lock")

    # Wait for any other process to finish downloading the file.
    # This way we can use the result from the current download without
    # spawning multiple concurrent downloads.
    while True:
        try:
            logger.debug(f"Attempting to acquire lock on {destination_lock}")
            with FileLock(str(destination_lock), timeout=5):
                logger.debug(f"Successfully acquired lock on {destination_lock}")
                return _download_to_or_read_from_cache(
                    url,
                    cache=cache,
                    dont_check_if_newer_than_seconds=dont_check_if_newer_than_seconds,
                )
        except Timeout:
            logger.warning(
                f"Failed to acquire lock on {destination_lock}, it is likely "
                f"being downloaded by another process. Retrying..."
            )


def _is_cached_file_fresh(
    destination: Path, dont_check_if_newer_than_seconds: float
) -> bool:
    """Check if a cached file exists and is fresh enough to use without checking.

    (In this context, "checking" means that later, beyond the scope of this function,
    we will query the server with an ETag to see if the file has changed or if we
    get a 304 Not Modified response.)

    A file is "fresh" if its age is positive and less than
    `dont_check_if_newer_than_seconds`.

    Returns True if the file is fresh, False otherwise.
    """
    if destination.is_file():
        age_seconds = get_age_seconds(destination)
        if age_seconds is None:
            raise RuntimeError(f"Error checking age of {destination}")
        if 0 <= age_seconds < dont_check_if_newer_than_seconds:
            logger.debug(
                f"Using cached file {destination} of age {age_seconds}s "
                f"without checking for updates"
            )
            return True
    return False


def _download_to_or_read_from_cache(
    url: str, *, cache: Path, dont_check_if_newer_than_seconds: float
) -> bytes:
    """Download a file to the cache directory and return the contents.

    This function is designed to be called from within a FileLock context to avoid
    multiple processes downloading the same file.

    If the file is newer than `dont_check_if_newer_than_seconds`, then immediately
    return the cached contents. Otherwise we pass the ETag from the last download
    in the headers to avoid downloading the file if it hasn't changed remotely.
    """
    destination = cache / cached_filename_for_url(url)
    destination_etag = destination.with_suffix(".etag")
    request_headers = {"User-Agent": "conda-lock"}
    # Return the contents immediately if the file is fresh
    if _is_cached_file_fresh(destination, dont_check_if_newer_than_seconds):
        return destination.read_bytes()
    # Add the ETag from the last download, if it exists, to the headers.
    # The ETag is used to avoid downloading the file if it hasn't changed remotely.
    # Otherwise, download the file and cache the contents and ETag.
    if destination.is_file() and destination_etag.is_file():
        old_etag = destination_etag.read_text().strip()
        request_headers["If-None-Match"] = old_etag
    # Download the file and cache the result.
    logger.debug(f"Requesting {url}")
    res = requests.get(url, headers=request_headers)
    if res.status_code == 304:
        logger.debug(f"{url} has not changed since last download, using {destination}")
    else:
        res.raise_for_status()
        destination.write_bytes(res.content)
        if "ETag" in res.headers:
            destination_etag.write_text(res.headers["ETag"])
        else:
            logger.warning("No ETag in response headers")
    logger.debug(f"Downloaded {url} to {destination}")
    return destination.read_bytes()


def cached_filename_for_url(url: str) -> str:
    """Return a filename for a URL that is probably unique to the URL.

    The filename is a 4-character hash of the URL, followed by the extension.
    If the extension is not alphanumeric or too long, it is omitted.

    >>> cached_filename_for_url("https://example.com/foo.json")
    'a5d7.json'
    >>> cached_filename_for_url("https://example.com/foo")
    '5ea6'
    >>> cached_filename_for_url("https://example.com/foo.bär")
    '2191'
    >>> cached_filename_for_url("https://example.com/foo.baaaaaar")
    '1861'
    """
    url_hash = hashlib.sha256(url.encode()).hexdigest()[:4]
    extension = url.split(".")[-1]
    if len(extension) <= 6 and re.match("^[a-zA-Z0-9]+$", extension):
        return f"{url_hash}.{extension}"
    else:
        return f"{url_hash}"


def clear_old_files_from_cache(cache: Path, *, max_age_seconds: float) -> None:
    """Remove files in the cache directory older than `max_age_seconds`.

    Also removes any files that somehow have a modification time in the future.

    For safety, this raises an error if `cache` is not a subdirectory of
    a directory named `"cache"`.
    """
    if not cache.parent.name == "cache":
        raise ValueError(
            f"Expected cache directory, got {cache}. Parent should be 'cache' ",
            f"not '{cache.parent.name}'",
        )
    for file in cache.iterdir():
        age_seconds = get_age_seconds(file)
        if age_seconds is None:
            # The file was probably already deleted.
            pass
        elif age_seconds < 0 or age_seconds >= max_age_seconds:
            try:
                file.unlink()
                logger.debug(f"Removed old cache file {file} of age {age_seconds}s")
            except FileNotFoundError:
                pass


def get_age_seconds(path: Path) -> Optional[float]:
    """Return the age of a file in seconds.

    On some filesystems, the age of a new file is sometimes slightly negative due to
    timestamp resolution. We treat negative ages within a small tolerance of zero
    as zero, which is important for the freshness check.
    """
    try:
        raw_age = datetime.now().timestamp() - path.stat().st_mtime
    except FileNotFoundError:
        return None
    if -TIMESTAMP_RESOLUTION_SECONDS <= raw_age < 0:
        return 0.0
    return raw_age



from conda_lock._vendor.conda.common.toposort import toposort
from conda_lock._vendor.conda.common.url import (
    mask_anaconda_token,
    split_anaconda_token,
)
from conda_lock._vendor.conda.models.match_spec import MatchSpec


__all__ = ["MatchSpec", "mask_anaconda_token", "split_anaconda_token", "toposort"]



import json
import logging
import time

from functools import cache
from pathlib import Path
from typing import TypedDict

from packaging.utils import NormalizedName
from packaging.utils import canonicalize_name as canonicalize_pypi_name

from conda_lock.lookup_cache import cached_download_file


logger = logging.getLogger(__name__)

DEFAULT_MAPPING_URL = "https://raw.githubusercontent.com/regro/cf-graph-countyfair/master/mappings/pypi/grayskull_pypi_mapping.json"


class MappingEntry(TypedDict):
    conda_name: str
    # legacy field, generally not used by anything anymore
    conda_forge: str
    pypi_name: NormalizedName


@cache
def _get_pypi_lookup(mapping_url: str) -> dict[NormalizedName, MappingEntry]:
    url = mapping_url
    if url.startswith("http://") or url.startswith("https://"):
        content = cached_download_file(url, cache_subdir_name="pypi-mapping")
    else:
        if url.startswith("file://"):
            path = url[len("file://") :]
        else:
            path = url
        content = Path(path).read_bytes()
    logger.debug("Parsing PyPI mapping")
    load_start = time.monotonic()
    if url.endswith(".json"):
        lookup = json.loads(content)
    else:
        import ruamel.yaml

        yaml = ruamel.yaml.YAML(typ="safe")
        lookup = yaml.load(content)
    load_duration = time.monotonic() - load_start
    logger.debug(f"Loaded {len(lookup)} entries in {load_duration:.2f}s")
    # lowercase and kebabcase the pypi names
    assert lookup is not None
    lookup = {canonicalize_pypi_name(k): v for k, v in lookup.items()}
    for v in lookup.values():
        v["pypi_name"] = canonicalize_pypi_name(v["pypi_name"])
    return lookup


def pypi_name_to_conda_name(name: str, mapping_url: str) -> str:
    """Convert a PyPI package name to a conda package name.

    >>> from conda_lock.lookup import DEFAULT_MAPPING_URL
    >>> pypi_name_to_conda_name("build", mapping_url=DEFAULT_MAPPING_URL)
    'python-build'

    >>> pypi_name_to_conda_name("zpfqzvrj", mapping_url=DEFAULT_MAPPING_URL)
    'zpfqzvrj'
    """
    cname = canonicalize_pypi_name(name)
    lookup = _get_pypi_lookup(mapping_url)
    if cname in lookup:
        entry = lookup[cname]
        res = entry.get("conda_name") or entry.get("conda_forge")
        if res is not None:
            return res

    logger.debug(f"Could not find conda name for {cname}. Assuming identity.")
    return cname


@cache
def _get_conda_lookup(mapping_url: str) -> dict[str, MappingEntry]:
    """
    Reverse grayskull name mapping to map conda names onto PyPI
    """
    return {
        record["conda_name"]: record
        for record in _get_pypi_lookup(mapping_url).values()
    }


def conda_name_to_pypi_name(name: str, mapping_url: str) -> NormalizedName:
    """return the pypi name for a conda package"""
    lookup = _get_conda_lookup(mapping_url=mapping_url)
    cname = canonicalize_pypi_name(name)
    return lookup.get(cname, {"pypi_name": cname})["pypi_name"]



from __future__ import annotations

from conda_lock._vendor.poetry.plugins.application_plugin import ApplicationPlugin
from conda_lock._vendor.poetry.plugins.plugin import Plugin


__all__ = ["ApplicationPlugin", "Plugin"]



"""Compute the content hash of a lock specification.

The content hash is used with `conda-lock --check-content-hash` to avoid unnecessary
relocking when the lock specification is unchanged.

Note that the content hash depends not only on the LockSpecification object but
also the virtual package specification.

WARNING: The fundamental concept of a content hash is seriously flawed:
<https://github.com/conda/conda-lock/issues/432#issuecomment-1637071282>

It is important to maintain the content hash for backwards compatibility,
but we should stop using it in the future.

Note that anything that modifies the content hash is a breaking change,
since it will invalidate all existing lockspecs.

The content hash is computed by JSON-serializing the lock specification and
the virtual package specification, and then hashing the serialized
representation.
"""

import hashlib
import json

from collections.abc import Sequence
from copy import deepcopy
from typing import Optional, Union, cast

from conda_lock.content_hash_types import (
    EmptyDict,
    HashableVirtualPackageRepresentation,
    PlatformSubdirStr,
    SerializedDependency,
    SerializedLockspec,
    SubdirMetadata,
)
from conda_lock.models.lock_spec import LockSpecification
from conda_lock.virtual_package import FakeRepoData


def compute_content_hashes(
    lock_spec: LockSpecification,
    virtual_package_repo: Optional[FakeRepoData],
    reinsert_spurious_build_number: bool = True,
    remove_new_nulls: bool = True,
) -> dict[PlatformSubdirStr, str]:
    """Compute the content hashes for the given lock specification.

    Args:
        lock_spec: The lock specification to compute the content hashes for.

        virtual_package_repo: The virtual package repository to use.
            If None, the content hash is computed without the VPR.

        reinsert_spurious_build_number: Whether to reinsert the spurious build
            number in the build string of the VPR. This prevents the content hash
            from changing when upgrading from v2 to v3.0.3.

        remove_new_nulls: Whether to remove newly added fields from the package
            specs when they are null. This prevents the content hash from changing
            when upgrading from v2 to v3.0.3.

    Returns:
        A dictionary of platform-specific content hashes.
    """
    result: dict[PlatformSubdirStr, str] = {}

    # This is done so that conda-lock >=3.0.3 will produce the same content
    # hashes as conda-lock v2.
    if reinsert_spurious_build_number and virtual_package_repo is not None:
        virtual_package_repo = _reinsert_spurious_build_number(virtual_package_repo)

    for platform in lock_spec.platforms:
        content = _content_for_platform(lock_spec, platform, virtual_package_repo)
        if remove_new_nulls:
            content = _remove_new_nulls(content)
        result[platform] = _dict_to_hash(content)
    return result


def _dict_to_json(
    d: Union[SerializedLockspec, HashableVirtualPackageRepresentation],
) -> str:
    """Produce a canonical JSON representation of the given dict."""
    return json.dumps(d, sort_keys=True)


def _json_to_hash(s: str) -> str:
    """Hash the given JSON string."""
    return hashlib.sha256(s.encode("utf-8")).hexdigest()


def _dict_to_hash(
    d: Union[SerializedLockspec, HashableVirtualPackageRepresentation],
) -> str:
    """Hash the given dict."""
    return _json_to_hash(_dict_to_json(d))


def _content_for_platform(
    lock_spec: LockSpecification,
    platform: PlatformSubdirStr,
    virtual_package_repo: Optional[FakeRepoData],
) -> SerializedLockspec:
    serialized_lockspec: SerializedLockspec = {
        "channels": [c.model_dump_json() for c in lock_spec.channels],
        "specs": [
            cast(SerializedDependency, p.model_dump())
            for p in sorted(
                lock_spec.dependencies[platform], key=lambda p: (p.manager, p.name)
            )
        ],
    }
    if lock_spec.pip_repositories:
        serialized_lockspec["pip_repositories"] = [
            repo.model_dump_json() for repo in lock_spec.pip_repositories
        ]
    if virtual_package_repo is not None:
        serialized_lockspec["virtual_package_hash"] = (
            _virtual_package_content_for_platform(virtual_package_repo, platform)
        )
    return serialized_lockspec


def _virtual_package_content_for_platform(
    virtual_package_repo: FakeRepoData,
    platform: PlatformSubdirStr,
) -> HashableVirtualPackageRepresentation:
    """Serialize the virtual package content into a dict for hashing.

    This is used in the computation of the content hash, and goes
    into the "virtual_package_hash" field of the serialized lockspec.
    """
    vpr_data = virtual_package_repo.all_repodata

    # We don't actually use these values! I'm including them to indicate
    # what I would have expected from the schema. See the code block
    # immediately below for the actual values.
    fallback_noarch: Union[SubdirMetadata, EmptyDict] = {
        "info": {"subdir": "noarch"},
        "packages": {},
    }
    fallback_platform: Union[SubdirMetadata, EmptyDict] = {
        "info": {"subdir": platform},
        "packages": {},
    }

    # It seems a bit of a schema violation, but the original implementation
    # did this, so we have to keep it in order to preserve consistency of
    # the hashes.
    fallback_noarch = {}
    fallback_platform = {}

    result: HashableVirtualPackageRepresentation = {
        "noarch": vpr_data.get("noarch", fallback_noarch),
        platform: vpr_data.get(platform, fallback_platform),
    }
    return result


def backwards_compatible_content_hashes(
    lock_spec: LockSpecification,
    virtual_package_repo: Optional[FakeRepoData],
    platform: PlatformSubdirStr,
) -> set[str]:
    """Compute a set of content hashes for equivalent lock specifications.

    Computing multiple content hashes allows us to support previous versions of
    the content hash computation for backwards compatibility.

    We could have adopted a more targeted strategy for producing specific variants
    of the VPR, but the VPR can also be customized, so it's hard to know exactly
    how it's constructed. Therefore we just enumerate all possible variants to be safe.

    Note that VPR=None is only used for old tests, and it corresponds to the case where
    VPR is unspecified rather than default. (TODO: replace those tests and eliminate
    this special case?)
    """
    virtual_package_repo_variants: list[FakeRepoData] = []
    if virtual_package_repo is not None:
        # Allow for equivalent legacy versions of the VPR to support
        # backwards compatibility so that we don't unnecessarily reject a good hash.
        # This list will be combinatorially expanded in the following steps.
        # (If the VPR is None, we leave the list empty, effectively skipping the
        # enumeration.)
        virtual_package_repo_variants = [virtual_package_repo]

    # Support both with and without the redundant __osx=10.15 package.
    for vpr in virtual_package_repo_variants.copy():
        if platform == "osx-64" and _contains_osx_11_0_0_tar_bz2(vpr):
            virtual_package_repo_variants.append(add_or_remove_osx_10_15_0_tar_bz2(vpr))

    # Reinsert spurious build number in build string
    for vpr in virtual_package_repo_variants.copy():
        virtual_package_repo_variants.append(_reinsert_spurious_build_number(vpr))

    # Compute virtual_package_repo parameter values to iterate over.
    vprs: Sequence[Optional[FakeRepoData]]
    if virtual_package_repo is None:
        assert len(virtual_package_repo_variants) == 0
        vprs = [None]
    else:
        assert len(virtual_package_repo_variants) > 0
        vprs = virtual_package_repo_variants

    # Compute the content hashes for the given lock specification and VPR variants.
    allowed_hashes: set[str] = set()
    for vpr_or_none in vprs:
        # We don't need to check cases involving reinserting the spurious build number
        # in the VPR since that's already covered by the VPR variants.
        # We do need to include both possible values of remove_new_nulls, because
        # that affects the package specs, not the VPR.
        allowed_hashes.add(
            compute_content_hashes(
                lock_spec=lock_spec,
                virtual_package_repo=vpr_or_none,
                reinsert_spurious_build_number=False,
                remove_new_nulls=False,
            )[platform]
        )
        allowed_hashes.add(
            compute_content_hashes(
                lock_spec=lock_spec,
                virtual_package_repo=vpr_or_none,
                reinsert_spurious_build_number=False,
                remove_new_nulls=True,
            )[platform]
        )
    return allowed_hashes


def add_or_remove_osx_10_15_0_tar_bz2(
    virtual_package_repo: FakeRepoData,
) -> FakeRepoData:
    """Add or remove the __osx 10.15 virtual package.

    Adds __osx 10.15 if it is not present, and removes it if it is present.
    This way, whichever convention we start with, the opposite convention will be
    produced.

    Rationale:
    In 6f69901 we started generating the default repodata based on
    default-virtual-packages.yaml instead of programmatically. But there was a bug
    in which we added both __osx 10.15 and 11.0. The extra 10.15 is ignored by conda
    and mamba, and 11.0 takes precedence. We added an option to readd the 10.15 package
    in 777dfbf.
    """
    result = virtual_package_repo.model_copy(deep=True)

    # We know this isn't empty because we only call this when 11.0 is present.
    rd: SubdirMetadata = cast(SubdirMetadata, result.all_repodata["osx-64"])

    packages = rd["packages"]
    if "__osx-10.15-0.tar.bz2" in packages:
        del packages["__osx-10.15-0.tar.bz2"]
    else:
        packages["__osx-10.15-0.tar.bz2"] = {
            "name": "__osx",
            "version": "10.15",
            "build_string": "",
            "build_number": 0,
            "noarch": "",
            "depends": [],
            "timestamp": 1577854800000,
            "package_type": "virtual_system",
            "build": "0",
            "subdir": "osx-64",
        }
    return result


def _contains_osx_11_0_0_tar_bz2(virtual_package_repo: FakeRepoData) -> bool:
    rd = virtual_package_repo.all_repodata.get("osx-64", {})
    if "packages" not in rd:
        return False
    rd = cast(SubdirMetadata, rd)
    return rd["packages"].get("__osx-11.0-0.tar.bz2") == {
        "name": "__osx",
        "version": "11.0",
        "build_string": "",
        "build_number": 0,
        "noarch": "",
        "depends": [],
        "timestamp": 1577854800000,
        "package_type": "virtual_system",
        "build": "0",
        "subdir": "osx-64",
    }


def _reinsert_spurious_build_number(virtual_package_repo: FakeRepoData) -> FakeRepoData:
    """Reinsert the spurious build number in the build string.

    This was introduced in v3.0.3 to reproduce the content hash of the v2 lockfiles.
    <https://github.com/conda/conda-lock/pull/776>

    Without the spurious build number:

    ```json
    "__archspec-1-x86_64.tar.bz2": {
        "build": "x86_64",
        "build_number": 0,
        "build_string": "x86_64",
        "depends": [],
        "name": "__archspec",
        "noarch": "",
        "package_type": "virtual_system",
        "subdir": "linux-64",
        "timestamp": 1577854800000,
        "version": "1"
    }
    ```

    With the spurious build number:
    ```json
    "__archspec-1-x86_64_0.tar.bz2": {
        "build": "x86_64_0",
        "build_number": 0,
        "build_string": "x86_64",
        "depends": [],
        "name": "__archspec",
        "noarch": "",
        "package_type": "virtual_system",
        "subdir": "linux-64",
        "timestamp": 1577854800000,
        "version": "1"
    }
    ```
    """
    result = virtual_package_repo.model_copy(deep=True)
    for platform in result.all_repodata:
        rd = result.all_repodata[platform]
        if "packages" in rd:
            rd = cast(SubdirMetadata, rd)
            for package_name, package_data in rd["packages"].copy().items():
                name = package_data["name"]
                version = package_data["version"]
                build_string = package_data["build_string"]
                build_number = package_data["build_number"]
                if len(build_string) > 0:
                    package_data["build"] = f"{build_string}_{build_number}"
                    new_name = f"{name}-{version}-{build_string}_{build_number}.tar.bz2"
                    rd["packages"][new_name] = package_data
                    del rd["packages"][package_name]
    return result


def _remove_new_nulls(content: SerializedLockspec) -> SerializedLockspec:
    """Remove newly added fields from the VPR when they are null.

    New fields added in v3.0.0 that are usually None but were absent in v2
    would alter the content hash, so we remove them for backwards compatibility.
    """
    result = deepcopy(content)
    for spec in result["specs"]:
        if "markers" in spec and spec["markers"] is None:
            del spec["markers"]
        if "subdirectory" in spec and spec["subdirectory"] is None:
            del spec["subdirectory"]
    return result



import pathlib
import typing

from typing import Optional, Union

from pydantic import BaseModel, Field, field_validator
from typing_extensions import Literal

from conda_lock.models import StrictModel
from conda_lock.models.channel import Channel
from conda_lock.models.pip_repository import PipRepository


class _BaseDependency(StrictModel):
    name: str
    manager: Literal["conda", "pip"] = "conda"
    category: str = "main"
    extras: list[str] = []
    markers: Optional[str] = None

    @field_validator("extras")
    @classmethod
    def sorted_extras(cls, v: list[str]) -> list[str]:
        return sorted(v)


class VersionedDependency(_BaseDependency):
    version: str
    build: Optional[str] = None
    conda_channel: Optional[str] = None
    hash: Optional[str] = None


class URLDependency(_BaseDependency):
    url: str
    hashes: list[str]


class VCSDependency(_BaseDependency):
    source: str
    vcs: str
    rev: Optional[str] = None
    subdirectory: Optional[str] = None


class PathDependency(_BaseDependency):
    path: str
    is_directory: bool
    subdirectory: Optional[str] = None


Dependency = Union[VersionedDependency, URLDependency, VCSDependency, PathDependency]


class Package(StrictModel):
    url: str
    hash: str


class PoetryMappedDependencySpec(StrictModel):
    url: Optional[str] = None
    manager: Literal["conda", "pip"]
    extras: list
    markers: Optional[str] = None
    poetry_version_spec: Optional[str] = None


class LockSpecification(BaseModel):
    dependencies: dict[str, list[Dependency]]
    # TODO: Should we store the auth info in here?
    channels: list[Channel]
    sources: list[pathlib.Path]
    pip_repositories: list[PipRepository] = Field(default_factory=list)
    allow_pypi_requests: bool = True

    @property
    def platforms(self) -> list[str]:
        return list(self.dependencies.keys())

    @field_validator("channels", mode="before")
    @classmethod
    def validate_channels(cls, v: list[Union[Channel, str]]) -> list[Channel]:
        for i, e in enumerate(v):
            if isinstance(e, str):
                e = Channel.from_string(e)
                v[i] = e
            if e.url == "nodefaults":
                raise ValueError("nodefaults channel is not allowed, ref #418")
        return typing.cast(list[Channel], v)

    @field_validator("pip_repositories", mode="before")
    @classmethod
    def validate_pip_repositories(
        cls, value: list[Union[PipRepository, str]]
    ) -> list[PipRepository]:
        for index, repository in enumerate(value):
            if isinstance(repository, str):
                value[index] = PipRepository.from_string(repository)
        return typing.cast(list[PipRepository], value)



import atexit
import json
import logging
import os
import pathlib

from collections import defaultdict
from collections.abc import Iterable
from importlib.resources import path
from types import TracebackType
from typing import (
    Literal,
    Optional,
    Union,
)

from pydantic import BaseModel, ConfigDict, Field, field_validator
from typing_extensions import TypeAlias

from conda_lock.content_hash_types import (
    HashableVirtualPackage,
    HashableVirtualPackageRepresentation,
    PackageNameStr,
    PlatformSubdirStr,
    SubdirMetadata,
)
from conda_lock.models.channel import Channel


logger = logging.getLogger(__name__)

# datetime.datetime(2020, 1, 1).timestamp()
DEFAULT_TIME = 1577854800000

VirtualPackageVersion: TypeAlias = str


with path("conda_lock", "default-virtual-packages.yaml") as p:
    DEFAULT_VIRTUAL_PACKAGES_YAML_PATH = p


class VirtualPackage(BaseModel):
    """A minimal representation of the required metadata for a virtual package.

    This is used by our specification. It's then converted to a FullVirtualPackage
    for computing the content hash. Then it's converted to a dict to be used in
    repodata.json.
    """

    model_config = ConfigDict(frozen=True)

    name: PackageNameStr
    version: VirtualPackageVersion
    build_string: str = ""

    def to_full_virtual_package(self) -> "FullVirtualPackage":
        return FullVirtualPackage(
            name=self.name,
            version=self.version,
            build_string=self.build_string,
        )

    def to_repodata_entry(
        self, *, subdir: PlatformSubdirStr
    ) -> tuple[str, HashableVirtualPackage]:
        p = self.to_full_virtual_package()
        out: HashableVirtualPackage = {
            "name": p.name,
            "version": p.version,
            "build_string": p.build_string,
            "build_number": p.build_number,
            "noarch": p.noarch,
            "depends": list(p.depends),
            "timestamp": p.timestamp,
            "package_type": p.package_type,
            "build": p.build,
            "subdir": subdir,
        }
        fname = f"{self.name}-{self.version}-{p.build}.tar.bz2"
        return fname, out


class FullVirtualPackage(VirtualPackage):
    """Everything necessary for repodata.json except subdir"""

    build_number: int = 0
    noarch: str = ""
    depends: tuple[str, ...] = Field(default_factory=tuple)
    timestamp: int = DEFAULT_TIME
    package_type: Optional[str] = "virtual_system"

    @property
    def build(self) -> str:
        if self.build_string:
            return self.build_string
        else:
            return str(self.build_number)


class FakeRepoData(BaseModel):
    base_path: pathlib.Path
    packages_by_subdir: defaultdict[FullVirtualPackage, set[PackageNameStr]] = Field(
        default_factory=lambda: defaultdict(set)  # type: ignore[arg-type,unused-ignore]
    )
    all_subdirs: set[PlatformSubdirStr] = {
        "noarch",
        "linux-aarch64",
        "linux-ppc64le",
        "linux-64",
        "osx-64",
        "osx-arm64",
        "win-64",
    }
    all_repodata: HashableVirtualPackageRepresentation = {}
    hash: Optional[str] = None
    old_env_vars: dict[str, Optional[str]] = {}

    @property
    def channel_url(self) -> str:
        if isinstance(self.base_path, pathlib.WindowsPath):
            return str(self.base_path.absolute())
        else:
            return f"file://{self.base_path.absolute().as_posix()}"

    @property
    def channel(self) -> Channel:
        # The URL is a file path, so there are no env vars. Thus we use the
        # raw Channel constructor here rather than the usual Channel.from_string().
        return Channel(url=self.channel_url, used_env_vars=())

    @property
    def channel_url_posix(self) -> str:
        if isinstance(self.base_path, pathlib.WindowsPath):
            # Mamba has a different return format for windows filepach urls and takes the form
            # file:///C:/dira/dirb
            return f"file:///{self.base_path.absolute().as_posix()}"
        else:
            return f"file://{self.base_path.absolute().as_posix()}"

    def add_package(
        self, package: VirtualPackage, subdirs: Iterable[PlatformSubdirStr] = ()
    ) -> None:
        subdirs = frozenset(subdirs)
        if not subdirs:
            subdirs = frozenset(["noarch"])
        self.packages_by_subdir[package.to_full_virtual_package()].update(subdirs)

    def _write_subdir(self, subdir: PlatformSubdirStr) -> SubdirMetadata:
        packages: dict[PackageNameStr, HashableVirtualPackage] = {}
        out: SubdirMetadata = {"info": {"subdir": subdir}, "packages": packages}
        for pkg, subdirs in self.packages_by_subdir.items():
            if subdir not in subdirs:
                continue
            fname, info_dict = pkg.to_repodata_entry(subdir=subdir)
            packages[fname] = info_dict

        (self.base_path / subdir).mkdir(exist_ok=True)
        content = json.dumps(out, sort_keys=True)
        (self.base_path / subdir / "repodata.json").write_text(content)
        return out

    def write(self) -> None:
        for subdirs in self.packages_by_subdir.values():
            self.all_subdirs.update(subdirs)

        for subdir in sorted(self.all_subdirs):
            repodata = self._write_subdir(subdir)
            self.all_repodata[subdir] = repodata

        logger.debug("Wrote fake repodata to %s", self.base_path)
        import glob

        for filename in glob.iglob(str(self.base_path / "**"), recursive=True):
            logger.debug(filename)
        logger.debug("repo: %s", self.channel_url)

    def __enter__(self) -> None:
        """Ensure that if glibc etc is set by the overrides we force the conda solver override variables"""
        env_vars_to_clear = set()
        for package in self.packages_by_subdir:
            if package.name.startswith("__"):
                upper_name = package.name.lstrip("_").upper()
                env_vars_to_clear.add(f"CONDA_OVERRIDE_{upper_name}")

        for e in env_vars_to_clear:
            self.old_env_vars[e] = os.environ.get(e)
            os.environ[e] = ""

    def __exit__(
        self,
        exc_type: Optional[type[BaseException]],
        exc: Optional[BaseException],
        traceback: Optional[TracebackType],
    ) -> None:
        """Clear out old vars"""
        for k, v in self.old_env_vars.items():
            if v is None:
                del os.environ[k]
            else:
                os.environ[k] = v


def _init_fake_repodata() -> FakeRepoData:
    import shutil
    import tempfile

    # tmp directory in github actions
    runner_tmp = os.environ.get("RUNNER_TEMP")
    tmp_dir = tempfile.mkdtemp(dir=runner_tmp)

    if not runner_tmp:
        # no need to bother cleaning up on CI
        def clean() -> None:
            shutil.rmtree(tmp_dir, ignore_errors=True)

        atexit.register(clean)

    tmp_path = pathlib.Path(tmp_dir)
    repodata = FakeRepoData(base_path=tmp_path)
    return repodata


def default_virtual_package_repodata(
    cuda_version: Union[Literal["default", ""], VirtualPackageVersion] = "default",
) -> FakeRepoData:
    """An empty cuda_version indicates that CUDA is unavailable."""
    """Define a reasonable modern set of virtual packages that should be safe enough to assume"""
    repodata = virtual_package_repo_from_specification(
        DEFAULT_VIRTUAL_PACKAGES_YAML_PATH,
        override_cuda_version=cuda_version,
        add_duplicate_osx_package=True,
    )
    return repodata


class VirtualPackageSpecSubdir(BaseModel):
    """Virtual packages for a specific subdir in a virtual-packages.yaml file"""

    packages: dict[PackageNameStr, VirtualPackageVersion]

    @field_validator("packages")
    @classmethod
    def validate_packages(
        cls, v: dict[PackageNameStr, VirtualPackageVersion]
    ) -> dict[PackageNameStr, VirtualPackageVersion]:
        for package_name in v:
            if not package_name.startswith("__"):
                raise ValueError(f"{package_name} is not a virtual package!")
        return v


class VirtualPackageSpec(BaseModel):
    """Virtual packages specified in a virtual-packages.yaml file"""

    subdirs: dict[PlatformSubdirStr, VirtualPackageSpecSubdir]


def _parse_virtual_package_spec(
    virtual_package_name: PackageNameStr, version_spec: str
) -> VirtualPackage:
    """Parse a virtual package specification into a VirtualPackage object.

    Args:
        virtual_package_name: The name of the virtual package (should start with '__')
        version_spec: The version specification string, optionally including a build string
            separated by a space

    Returns:
        A VirtualPackage object with the parsed name, version, and build string

    Examples:
        >>> _parse_virtual_package_spec("__unix", "0")
        VirtualPackage(name='__unix', version='0', build_string='')
        >>> _parse_virtual_package_spec("__archspec", "1 x86_64")
        VirtualPackage(name='__archspec', version='1', build_string='x86_64')
    """
    version_parts = version_spec.split(" ", 1)
    assert len(version_parts) in (1, 2)
    if len(version_parts) == 1:
        parsed_version, build_string = version_spec, ""
    else:
        parsed_version, build_string = version_parts
    return VirtualPackage(
        name=virtual_package_name,
        version=parsed_version,
        build_string=build_string,
    )


def virtual_package_repo_from_specification(
    virtual_package_spec_file: pathlib.Path,
    add_duplicate_osx_package: bool = False,
    override_cuda_version: Union[
        Literal["default", ""], VirtualPackageVersion
    ] = "default",
) -> FakeRepoData:
    import yaml

    with virtual_package_spec_file.open("r") as fp:
        data = yaml.safe_load(fp)
    logging.debug("Virtual package spec: %s", data)

    virtual_package_spec = VirtualPackageSpec.model_validate(data)

    repodata = _init_fake_repodata()
    for subdir, subdir_spec in virtual_package_spec.subdirs.items():
        for virtual_package_name, version_spec in subdir_spec.packages.items():
            # Override the CUDA version if specified.
            if virtual_package_name == "__cuda" and override_cuda_version != "default":
                if override_cuda_version == "":
                    continue
                version_spec = override_cuda_version

            virtual_package = _parse_virtual_package_spec(
                virtual_package_name, version_spec
            )
            repodata.add_package(virtual_package, subdirs=[subdir])

    if add_duplicate_osx_package:
        # This is to preserve exact consistency with previous versions of conda-lock.
        # Previous versions of conda-lock would add the __osx package twice, once with
        # version "10.15" and once with version "11.0". The package with version "10.15"
        # is ignored by conda and mamba, but is still present in the virtual repodata,
        # and contributes to the content hash.
        # <https://github.com/conda/conda-lock/blob/f5323e7a71259ed17173401ec4cd728c6d161fe1/conda_lock/virtual_package.py#L191-L252>
        package = VirtualPackage(name="__osx", version="10.15")
        repodata.add_package(package, subdirs=["osx-64"])

    repodata.write()
    return repodata


if __name__ == "__main__":
    logging.basicConfig(level=logging.DEBUG)
    fil = (
        pathlib.Path(__file__).parent.parent
        / "tests"
        / "test-cuda"
        / "virtual-packages.yaml"
    )
    rd = virtual_package_repo_from_specification(fil)
    print(rd)
    print((rd.base_path / "linux-64" / "repodata.json").read_text())



"""This module contains the `unify_platform_independent_deps` function.

This is deliberately placed in a separate module since it doesn't directly have to
do with the logic for exporting a lock specification, and the logic is a bit involved.
"""

from collections import defaultdict
from typing import NamedTuple, Optional, Union

from conda_lock.models.lock_spec import Dependency


class EditableDependency(NamedTuple):
    name: str
    path: str


class DepKey(NamedTuple):
    """A hashable key for a `Dependency` object and the platform under consideration.

    There are two main motivations for this class.

    1. A `Dependency` does not have its own `platform` attribute, but we need to
    categorize dependencies by platform. This class allows us to do that.

    2. A `Dependency` is a Pydantic model, so it isn't hashable and can't be used as a
    key in a dictionary. This class is hashable and can be used as a key, enabling
    us to index dependencies by their attributes.

    When `platform` is `None`, this signifies a platform-independent dependency.
    """

    name: str
    category: str
    platform: Optional[str]
    manager: str

    def drop_platform(self) -> "DepKey":
        return self._replace(platform=None)


def unify_platform_independent_deps(
    dependencies: dict[str, list[Dependency]],
    *,
    editables: Optional[list[EditableDependency]] = None,
) -> dict[DepKey, Union[Dependency, EditableDependency]]:
    """Combine identical dependencies for all platforms into a single dependency.

    Returns a tuple of two dictionaries:

    >>> from conda_lock.models.lock_spec import VersionedDependency
    >>> numpy1a = VersionedDependency(name="numpy", version="1.2.3")
    >>> numpy1b = VersionedDependency(name="numpy", version="1.2.3")
    >>> pandas1 = VersionedDependency(name="pandas", version="4.5.6")
    >>> pandas2 = VersionedDependency(name="pandas", version="7.8.9")
    >>> xarray = VersionedDependency(name="xarray", version="1.2.3")
    >>> dependencies = {
    ...     "linux-64": [numpy1a, pandas1],
    ...     "osx-64": [numpy1b, pandas1, xarray],
    ...     "win-64": [numpy1a, pandas2],
    ... }

    Since `numpy1a` and `numpy1b` are equal, `numpy` is platform-independent.

    `xarray` only appears on `osx-64`.

    `pandas` is present on all platforms, but the versions aren't all the same.

    >>> unified = unify_platform_independent_deps(dependencies)
    >>> for key in unified:
    ...     print(key)
    DepKey(name='numpy', category='main', platform=None, manager='conda')
    DepKey(name='pandas', category='main', platform='linux-64', manager='conda')
    DepKey(name='pandas', category='main', platform='osx-64', manager='conda')
    DepKey(name='xarray', category='main', platform='osx-64', manager='conda')
    DepKey(name='pandas', category='main', platform='win-64', manager='conda')

    The full result:

    >>> unified  # doctest: +NORMALIZE_WHITESPACE
    {DepKey(name='numpy', category='main', platform=None, manager='conda'):
      VersionedDependency(name='numpy', manager='conda', category='main', extras=[],
        markers=None, version='1.2.3', build=None, conda_channel=None, hash=None),
     DepKey(name='pandas', category='main', platform='linux-64', manager='conda'):
      VersionedDependency(name='pandas', manager='conda', category='main', extras=[],
        markers=None, version='4.5.6', build=None, conda_channel=None, hash=None),
     DepKey(name='pandas', category='main', platform='osx-64', manager='conda'):
      VersionedDependency(name='pandas', manager='conda', category='main', extras=[],
        markers=None, version='4.5.6', build=None, conda_channel=None, hash=None),
     DepKey(name='xarray', category='main', platform='osx-64', manager='conda'):
      VersionedDependency(name='xarray', manager='conda', category='main', extras=[],
        markers=None, version='1.2.3', build=None, conda_channel=None, hash=None),
     DepKey(name='pandas', category='main', platform='win-64', manager='conda'):
      VersionedDependency(name='pandas', manager='conda', category='main', extras=[],
        markers=None, version='7.8.9', build=None, conda_channel=None, hash=None)}
    """
    indexed_deps: dict[DepKey, Dependency] = {}
    for platform, deps in dependencies.items():
        for dep in deps:
            key = DepKey(
                name=dep.name,
                category=dep.category,
                platform=platform,
                manager=dep.manager,
            )
            if key in indexed_deps:
                raise ValueError(
                    f"Duplicate dependency {key}: {dep}, {indexed_deps[key]}"
                )
            # In the beginning each dep has a platform.
            assert key.platform is not None
            indexed_deps[key] = dep

    # Collect deps which differ only by platform
    collected_deps: dict[DepKey, list[Dependency]] = defaultdict(list)
    for key, dep in indexed_deps.items():
        collected_deps[key.drop_platform()].append(dep)

    editable_deps: dict[DepKey, EditableDependency] = {}
    for editable in editables or []:
        key = DepKey(name=editable.name, category="main", platform=None, manager="pip")
        if key in collected_deps:
            raise ValueError(
                f"Editable dependency {editable.name} conflicts with existing "
                f"dependency {collected_deps[key][0]}"
            )
        editable_deps[key] = editable

    # Check for platform-independent dependencies
    num_platforms = len(dependencies.keys())
    platform_independent_deps: dict[DepKey, Dependency] = {
        np_key: deps[0]
        for np_key, deps in collected_deps.items()
        # It's independent if there's a dep for each platform and they're all the same.
        if len(deps) == num_platforms
        and all(curr == next for curr, next in zip(deps, deps[1:]))
    }
    assert all(key.platform is None for key in platform_independent_deps)

    # The platform-specific dependencies are now those not in platform_independent_deps.
    platform_specific_deps: dict[DepKey, Dependency] = {
        key: dep
        for key, dep in indexed_deps.items()
        if key.drop_platform() not in platform_independent_deps
    }
    assert all(key.platform is not None for key in platform_specific_deps)

    combined = {**platform_independent_deps, **editable_deps, **platform_specific_deps}
    return combined
